from meerkat.columns.abstract import AbstractColumn
from meerkat.writers.abstract import AbstractWriter


class ConcatWriter(AbstractWriter):
    def __init__(
        self,
        output_type: type = AbstractColumn,
        template: AbstractColumn = None,
        *args,
        **kwargs,
    ):
        super(ConcatWriter, self).__init__(*args, **kwargs)
        self.output_type = output_type
        self.template = template

    def open(self) -> None:
        self.outputs = []

    def write(self, data, **kwargs) -> None:
        # convert to Meerkat column if not already
        if self.template is not None:
            if isinstance(data, AbstractColumn):
                data = data.data
            data = self.template._clone(data=data)
        elif not isinstance(data, AbstractColumn):
            data = self.output_type(data)

        self.outputs.append(data)

    def flush(self):
        pass

    def close(self, *args, **kwargs):
        pass

    def finalize(self, *args, **kwargs) -> None:
        from meerkat.ops.concat import concat

        return concat(self.outputs)



import json
import os
from typing import Dict, Mapping

from tqdm import tqdm

import meerkat as mk


def crop_object(row: Mapping[str, object]):
    img = row["image"]

    length = max(row["h"], row["w"])
    box = (
        max(row["x"] - ((length - row["w"]) / 2), 0),
        max(row["y"] - ((length - row["h"]) / 2), 0),
        min(row["x"] + row["w"] + ((length - row["w"]) / 2), img.width),
        min(row["y"] + row["h"] + ((length - row["h"]) / 2), img.height),
    )
    return img.crop(box)


def build_gqa_dps(dataset_dir: str, write: bool = False) -> Dict[str, mk.DataPanel]:

    objects = []
    images = []
    relations = []
    attributes = []
    for split in ["train", "val"]:
        print(f"Loading {split} scene graphs...")
        with open(os.path.join(dataset_dir, f"{split}_sceneGraphs.json")) as f:
            graphs = json.load(f)
        for image_id, graph in tqdm(graphs.items()):
            image_id = int(image_id)  # convert to int for faster filtering and joins
            for object_id, obj in graph.pop("objects").items():
                object_id = int(
                    object_id
                )  # convert to int for faster filtering and joins
                for relation in obj.pop("relations"):
                    relations.append(
                        {
                            "subject_object_id": object_id,
                            "object_id": int(relation["object"]),
                            "name": relation["name"],
                        }
                    )
                for attribute in obj.pop("attributes"):
                    attributes.append({"object_id": object_id, "attribute": attribute})
                objects.append({"object_id": object_id, "image_id": image_id, **obj})
            images.append({"image_id": image_id, **graph})

    # prepare DataPanels
    print("Preparing DataPanels...")
    image_dp = mk.DataPanel(images)
    image_dp["image"] = mk.ImageColumn(
        image_dp["image_id"].map(
            lambda x: os.path.join(dataset_dir, "images", f"{x}.jpg")
        )
    )
    object_dp = mk.DataPanel(objects).merge(
        image_dp[["image_id", "image", "height", "width"]], on="image_id"
    )
    object_dp["object_image"] = object_dp.to_lambda(crop_object)
    # filter out objects with no width or height
    object_dp = object_dp.lz[(object_dp["h"] != 0) & (object_dp["w"] != 0)]
    # filter out objects whose bounding boxes are not contained within the image
    object_dp = object_dp.lz[
        (object_dp["x"] < object_dp["width"]) & (object_dp["y"] < object_dp["height"])
    ]

    dps = {
        "images": image_dp,
        "objects": object_dp,
        "relations": mk.DataPanel(relations),
        "attributes": mk.DataPanel(attributes),
    }

    if write:
        write_gqa_dps(dps=dps, dataset_dir=dataset_dir)
    return dps


def read_gqa_dps(dataset_dir: str) -> Dict[str, mk.DataPanel]:
    return {
        key: mk.DataPanel.read(os.path.join(dataset_dir, f"{key}.mk"))
        for key in ["attributes", "relations", "objects", "images"]
    }


def write_gqa_dps(dps: Mapping[str, mk.DataPanel], dataset_dir: str):
    for key, dp in dps.items():
        dp.write(os.path.join(dataset_dir, f"{key}.mk"))



from __future__ import annotations

import abc


class AbstractCell(abc.ABC):
    def __init__(self, *args, **kwargs):
        super(AbstractCell, self).__init__(*args, **kwargs)

    def get(self, *args, **kwargs) -> object:
        """Get me the thing that this cell exists for."""
        raise NotImplementedError("Must implement `get`.")

    @property
    def metadata(self) -> dict:
        """Get the metadata associated with this cell."""
        return {}

    def loader(self, *args, **kwargs) -> object:
        return self

    def __getitem__(self, index):
        return self.get()[index]

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def __eq__(self, other):
        raise NotImplementedError



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Hashable, Sequence, Tuple, Union

import pandas as pd
import torch

from meerkat.block.ref import BlockRef
from meerkat.columns.numpy_column import NumpyArrayColumn
from meerkat.columns.tensor_column import TensorColumn

from .abstract import AbstractBlock, BlockIndex, BlockView


class PandasBlock(AbstractBlock):
    @dataclass(eq=True, frozen=True)
    class Signature:
        nrows: int
        klass: type

    def __init__(self, data: pd.DataFrame, *args, **kwargs):
        super(PandasBlock, self).__init__(*args, **kwargs)
        self.data = data

    @property
    def signature(self) -> Hashable:
        return self.Signature(
            klass=PandasBlock,
            # we don't
            nrows=len(self.data),
        )

    def _get_data(self, index: BlockIndex) -> pd.Series:
        return self.data[index]

    @classmethod
    def from_column_data(cls, data: pd.Series) -> Tuple[PandasBlock, BlockView]:
        """[summary]

        Args:
            data (np.ndarray): [description]
            names (Sequence[str]): [description]

        Raises:
            ValueError: [description]

        Returns:
            Tuple[PandasBlock, Mapping[str, BlockIndex]]: [description]
        """
        data = pd.DataFrame({"col": data})
        block = cls(data)
        return BlockView(block_index="col", block=block)

    @classmethod
    def _consolidate(
        cls,
        block_refs: Sequence[BlockRef],
    ) -> BlockRef:
        df = pd.DataFrame(
            # need to ignore index when concatenating
            {
                name: ref.block.data[col._block_index].reset_index(drop=True)
                for ref in block_refs
                for name, col in ref.items()
            }
        )
        block = cls(df)

        # pull out the block columns from all the block_refs
        columns = {}
        for ref in block_refs:
            columns.update(ref)

        new_columns = {
            name: col._clone(data=block[name]) for name, col in columns.items()
        }
        return BlockRef(block=block, columns=new_columns)

    @staticmethod
    def _convert_index(index):
        if torch.is_tensor(index):
            # need to convert to numpy for boolean indexing
            return index.numpy()
        if isinstance(index, NumpyArrayColumn):
            return index.data
        if isinstance(index, TensorColumn):
            # need to convert to numpy for boolean indexing
            return index.data.numpy()
        if isinstance(index, pd.Series):
            # need to convert to numpy for boolean indexing
            return index.values
        from meerkat.columns.pandas_column import PandasSeriesColumn

        if isinstance(index, PandasSeriesColumn):
            return index.data.values
        return index

    def _get(
        self, index, block_ref: BlockRef, materialize: bool = True
    ) -> Union[BlockRef, dict]:
        index = self._convert_index(index)
        # TODO: check if they're trying to index more than just the row dimension
        data = self.data.iloc[index]
        if isinstance(index, int):
            # if indexing a single row, we do not return a block manager, just a dict
            return {
                name: data[col._block_index] for name, col in block_ref.columns.items()
            }
        block = self.__class__(data)

        columns = {
            name: col._clone(data=block[col._block_index])
            for name, col in block_ref.columns.items()
        }
        # note that the new block may share memory with the old block
        return BlockRef(block=block, columns=columns)

    def _write_data(self, path: str):
        self.data.reset_index(drop=True).to_feather(os.path.join(path, "data.feather"))

    @staticmethod
    def _read_data(path: str, mmap: bool = False):
        return pd.read_feather(os.path.join(path, "data.feather"))



from collections import defaultdict
from typing import Callable, Dict, List, Optional, Sequence

import yaml
from yaml.constructor import ConstructorError


class MeerkatLoader(yaml.FullLoader):
    """PyYaml does not load unimported modules for safety reasons. We want to allow
    importing only meerkat modules
    """

    def find_python_module(self, name: str, mark, unsafe=False):
        try:
            return super().find_python_module(name=name, mark=mark, unsafe=unsafe)
        except ConstructorError as e:
            if name.startswith("meerkat."):
                __import__(name)
            else:
                raise e
            return super().find_python_module(name=name, mark=mark, unsafe=unsafe)

    def find_python_name(self, name, mark, unsafe=False):
        if "." in name:
            module_name, _ = name.rsplit(".", 1)
        else:
            module_name = "builtins"

        try:
            return super().find_python_name(name=name, mark=mark, unsafe=unsafe)
        except ConstructorError as e:
            if name.startswith("meerkat."):
                __import__(module_name)
            else:
                raise e
            return super().find_python_name(name=name, mark=mark, unsafe=unsafe)


def convert_to_batch_column_fn(
    function: Callable, with_indices: bool, materialize: bool = True, **kwargs
):
    """Batch a function that applies to an example."""

    def _function(batch: Sequence, indices: Optional[List[int]], *args, **kwargs):
        # Pull out the batch size
        batch_size = len(batch)

        # Iterate and apply the function
        outputs = None
        for i in range(batch_size):

            # Apply the unbatched function
            if with_indices:
                output = function(
                    batch[i] if materialize else batch.lz[i],
                    indices[i],
                    *args,
                    **kwargs,
                )
            else:
                output = function(
                    batch[i] if materialize else batch.lz[i],
                    *args,
                    **kwargs,
                )

            if i == 0:
                # Create an empty dict or list for the outputs
                outputs = defaultdict(list) if isinstance(output, dict) else []

            # Append the output
            if isinstance(output, dict):
                for k in output.keys():
                    outputs[k].append(output[k])
            else:
                outputs.append(output)

        if isinstance(outputs, dict):
            return dict(outputs)
        return outputs

    if with_indices:
        # Just return the function as is
        return _function
    else:
        # Wrap in a lambda to apply the indices argument
        return lambda batch, *args, **kwargs: _function(batch, None, *args, **kwargs)


def convert_to_batch_fn(
    function: Callable, with_indices: bool, materialize: bool = True, **kwargs
):
    """Batch a function that applies to an example."""

    def _function(
        batch: Dict[str, List], indices: Optional[List[int]], *args, **kwargs
    ):
        # Pull out the batch size
        batch_size = len(batch[list(batch.keys())[0]])

        # Iterate and apply the function
        outputs = None
        for i in range(batch_size):

            # Apply the unbatched function
            if with_indices:
                output = function(
                    {k: v[i] if materialize else v.lz[i] for k, v in batch.items()},
                    indices[i],
                    *args,
                    **kwargs,
                )
            else:
                output = function(
                    {k: v[i] if materialize else v.lz[i] for k, v in batch.items()},
                    *args,
                    **kwargs,
                )

            if i == 0:
                # Create an empty dict or list for the outputs
                outputs = defaultdict(list) if isinstance(output, dict) else []

            # Append the output
            if isinstance(output, dict):
                for k in output.keys():
                    outputs[k].append(output[k])
            else:
                outputs.append(output)

        if isinstance(outputs, dict):
            return dict(outputs)
        return outputs

    if with_indices:
        # Just return the function as is
        return _function
    else:
        # Wrap in a lambda to apply the indices argument
        return lambda batch, *args, **kwargs: _function(batch, None, *args, **kwargs)



"""Import ml module classes."""
# flake8: noqa
import warnings

from meerkat.errors import ExperimentalWarning

warnings.warn(
    ExperimentalWarning(
        "The `meerkat.ml` module is experimental and has limited test coverage. "
        "Proceed with caution."
    )
)
from meerkat.ml.activation import ActivationOp
from meerkat.ml.callbacks import ActivationCallback, load_activations
from meerkat.ml.embedding_column import EmbeddingColumn
from meerkat.ml.huggingfacemodel import HuggingfaceModel
from meerkat.ml.instances_column import InstancesColumn
from meerkat.ml.prediction_column import ClassificationOutputColumn
from meerkat.ml.segmentation_column import SegmentationOutputColumn
from meerkat.ml.tensormodel import TensorModel



import collections.abc
from typing import List, Sequence, Union

import numpy as np

from meerkat import DataPanel, ListColumn
from meerkat.columns.cell_column import CellColumn
from meerkat.columns.lambda_column import LambdaColumn
from meerkat.columns.numpy_column import NumpyArrayColumn
from meerkat.columns.tensor_column import TensorColumn
from meerkat.errors import MergeError
from meerkat.provenance import capture_provenance


@capture_provenance(capture_args=["left_on", "on", "right_on", "how"])
def merge(
    left: DataPanel,
    right: DataPanel,
    how: str = "inner",
    on: Union[str, List[str]] = None,
    left_on: Union[str, List[str]] = None,
    right_on: Union[str, List[str]] = None,
    sort: bool = False,
    suffixes: Sequence[str] = ("_x", "_y"),
    validate=None,
    keep_indexes: bool = False,
):
    if how == "cross":
        raise ValueError("DataPanel does not support cross merges.")  # pragma: no cover

    if (on is None) and (left_on is None) and (right_on is None):
        raise MergeError("Merge expects either `on` or `left_on` and `right_on`")

    left_on = on if left_on is None else left_on
    right_on = on if right_on is None else right_on
    # cast `left_on` and `right_on` to lists
    left_on = [left_on] if isinstance(left_on, str) else left_on
    right_on = [right_on] if isinstance(right_on, str) else right_on

    # ensure we can merge on specified columns
    _check_merge_columns(left, left_on)
    _check_merge_columns(right, right_on)

    # convert datapanels to dataframes so we can apply Pandas merge
    # (1) only include columns we are joining on
    left_df = left[left_on].to_pandas()
    right_df = right[right_on].to_pandas()
    # (2) add index columns, which we'll use to reconstruct the columns we excluded from
    # the Pandas merge
    if ("__right_indices__" in right_df) or ("__left_indices__" in left_df):
        raise MergeError(
            "The column names '__right_indices__' and '__left_indices__' cannot appear "
            "in the right and left panels respectively. They are used by merge."
        )
    left_df["__left_indices__"] = np.arange(len(left_df))
    right_df["__right_indices__"] = np.arange(len(right_df))

    # apply pandas merge
    merged_df = left_df.merge(
        right_df,
        how=how,
        left_on=left_on,
        right_on=right_on,
        sort=sort,
        validate=validate,
        suffixes=suffixes,
    )
    left_indices = merged_df.pop("__left_indices__").values
    right_indices = merged_df.pop("__right_indices__").values
    merged_df = merged_df[set(left_on) & set(right_on)]

    # reconstruct other columns not in the `left_on & right_on` using `left_indices`
    # and `right_indices`, the row order returned by merge
    def _cols_to_construct(dp: DataPanel):
        # don't construct columns in both `left_on` and `right_on` because we use
        # `merged_df` for these
        return [k for k in dp.keys() if k not in (set(left_on) & set(right_on))]

    new_left = _construct_from_indices(left[_cols_to_construct(left)], left_indices)
    new_right = _construct_from_indices(right[_cols_to_construct(right)], right_indices)

    # concatenate the two new datapanels
    merged_dp = new_left.append(new_right, axis="columns", suffixes=suffixes)

    # add columns in both `left_on` and `right_on`, casting to the column type in left
    for name, column in merged_df.iteritems():
        merged_dp.add_column(name, left[name]._clone(data=column.values))
        merged_dp.data.reorder(merged_dp.columns[-1:] + merged_dp.columns[:-1])
    if (
        not keep_indexes
        and ("index" + suffixes[0]) in merged_dp
        and ("index" + suffixes[1]) in merged_dp
    ):
        merged_dp.remove_column("index" + suffixes[0])
        merged_dp.remove_column("index" + suffixes[1])

    return merged_dp


def _construct_from_indices(dp: DataPanel, indices: np.ndarray):
    if np.isnan(indices).any():
        # when performing "outer", "left", and "right" merges, column indices output
        # by pandas merge can include `nan` in rows corresponding to merge keys that
        # only appear in one of the two panels. For these columns, we convert the
        # column to  ListColumn, and fill with "None" wherever indices is "nan".
        data = {
            name: ListColumn(
                [None if np.isnan(index) else col.lz[int(index)] for index in indices]
            )
            for name, col in dp.items()
        }
        return dp._clone(data=data)
    else:
        # if there are no `nan`s in the indices, then we can just lazy index the
        # original column
        return dp.lz[indices]


def _check_merge_columns(dp: DataPanel, on: List[str]):
    for name in on:
        column = dp[name]
        if isinstance(column, NumpyArrayColumn) or isinstance(column, TensorColumn):
            if len(column.shape) > 1:
                raise MergeError(
                    f"Cannot merge on column `{name}`, has more than one dimension."
                )
        elif isinstance(column, ListColumn):
            if not all(
                [isinstance(cell, collections.abc.Hashable) for cell in column.lz]
            ):
                raise MergeError(
                    f"Cannot merge on column `{name}`, contains unhashable objects."
                )

        elif isinstance(column, CellColumn) or isinstance(column, LambdaColumn):
            if not all(
                [isinstance(cell, collections.abc.Hashable) for cell in column.lz]
            ):
                raise MergeError(
                    f"Cannot merge on column `{name}`, contains unhashable cells."
                )



import os


class stderr_suppress(object):
    """A context manager for doing a "deep suppression" of stdout and stderr in
    Python.

    This is necessary when reading in a corrupted video, or else stderr
    will emit 10000s of errors via ffmpeg. Great for decoding IRL, not
    great for loading 100s of corrupted videos.
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fd = os.open(os.devnull, os.O_RDWR)
        # Save stderr (2) file descriptor.
        self.save_fd = os.dup(2)

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fd, 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fd, 2)
        # Close all file descriptors
        os.close(self.null_fd)
        os.close(self.save_fd)



from __future__ import annotations

import logging
from collections.abc import Sequence
from types import SimpleNamespace
from typing import Union

import numpy as np
import pandas as pd
import torch

from meerkat.columns.tensor_column import TensorColumn
from meerkat.tools.lazy_loader import LazyLoader

faiss = LazyLoader("faiss")
umap = LazyLoader("umap")
umap_plot = LazyLoader("umap.plot")
sklearn_decom = LazyLoader("sklearn.decomposition")

Columnable = Union[Sequence, np.ndarray, pd.Series, torch.Tensor]

logger = logging.getLogger(__name__)


class EmbeddingColumn(TensorColumn):
    def __init__(
        self,
        data: Sequence = None,
        *args,
        **kwargs,
    ):
        if data is not None and isinstance(data, TensorColumn):
            data = data._data
        super(EmbeddingColumn, self).__init__(data=data, *args, **kwargs)

        # Cast to float32
        self._data = self._data.type(torch.FloatTensor)

        self.faiss_index = None

    def build_faiss_index(self, index=None, overwrite=False):
        if self.ndim < 2:
            raise ValueError("Building an index requires `ndim` >= 2.")

        if self.faiss_index is not None and not overwrite:
            return

        if index is None:
            index = faiss.IndexFlatL2

        # Create the faiss index
        self.faiss_index = index(self.shape[1])

        # Add the data: must be np.ndarray
        self.faiss_index.add(self.numpy())

    def search(self, query, k: int):
        assert (
            self.faiss_index is not None
        ), "You must call ``build_faiss_index`` first."
        if isinstance(query, np.ndarray):
            query = query.astype("float32")
            return self.faiss_index.search(query, k)
        return NotImplemented("`query` be np.ndarray.")

    def pca(self, n_components=2):
        pca = sklearn_decom.PCA(n_components)
        return SimpleNamespace(
            embeddings=pca.fit_transform(self.numpy()),
            reducer=pca,
        )

    def umap(self, n_neighbors=15, n_components=2):
        reducer = umap.UMAP(
            n_neighbors=n_neighbors,
            n_components=n_components,
        )
        return SimpleNamespace(
            embeddings=reducer.fit_transform(self.numpy()),
            reducer=reducer,
        )

    def visualize_umap(self, n_neighbors=15, n_components=2, point_size=4):
        p = umap_plot.interactive(
            self.umap(n_neighbors, n_components).reducer,
            point_size=point_size,
        )
        umap_plot.show(p)



"""DataPanel class."""
from __future__ import annotations

import logging
import os
import pathlib
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import cytoolz as tz
import datasets
import dill
import numpy as np
import pandas as pd
import pyarrow as pa
import torch
import ujson as json
import yaml
from jsonlines import jsonlines
from pandas._libs import lib

import meerkat
from meerkat.block.manager import BlockManager
from meerkat.columns.abstract import AbstractColumn
from meerkat.columns.cell_column import CellColumn
from meerkat.mixins.cloneable import CloneableMixin
from meerkat.mixins.inspect_fn import FunctionInspectorMixin
from meerkat.mixins.lambdable import LambdaMixin
from meerkat.mixins.mapping import MappableMixin
from meerkat.mixins.materialize import MaterializationMixin
from meerkat.provenance import ProvenanceMixin, capture_provenance
from meerkat.tools.utils import MeerkatLoader, convert_to_batch_fn

logger = logging.getLogger(__name__)

Example = Dict
Batch = Dict[str, Union[List, AbstractColumn]]
BatchOrDataset = Union[Batch, "DataPanel"]


class DataPanel(
    CloneableMixin,
    FunctionInspectorMixin,
    LambdaMixin,
    MappableMixin,
    MaterializationMixin,
    ProvenanceMixin,
):
    """Meerkat DataPanel class."""

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "meerkat/"

    # Create a directory
    logdir.mkdir(parents=True, exist_ok=True)

    def __init__(
        self,
        data: Union[dict, list, datasets.Dataset] = None,
        *args,
        **kwargs,
    ):
        super(DataPanel, self).__init__(
            *args,
            **kwargs,
        )
        logger.debug("Creating DataPanel.")

        self.data = data

        # TODO(Sabri): fix add_index for new datset
        # Add an index to the dataset
        if not self.has_index:
            self._add_index()

    def _repr_pandas_(self, max_rows: int = None):
        if max_rows is None:
            max_rows = meerkat.config.DisplayOptions.max_rows

        df, formatters = self.data._repr_pandas_(max_rows=max_rows)
        rename = {k: f"{k} ({v.__class__.__name__})" for k, v in self.items()}
        return (
            df[self.columns].rename(columns=rename),
            {rename[k]: v for k, v in formatters.items()},
        )

    def _repr_html_(self, max_rows: int = None):
        if max_rows is None:
            max_rows = meerkat.config.DisplayOptions.max_rows

        df, formatters = self._repr_pandas_(max_rows=max_rows)

        return df.to_html(formatters=formatters, max_rows=max_rows, escape=False)

    def streamlit(self):
        return self._repr_pandas_()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}" f"(nrows: {self.nrows}, ncols: {self.ncols})"
        )

    def __len__(self):
        return self.nrows

    def __contains__(self, item):
        return item in self.columns

    @property
    def data(self) -> BlockManager:
        """Get the underlying data (excluding invisible rows).

        To access underlying data with invisible rows, use `_data`.
        """
        return self._data

    def _set_data(self, value: Union[BlockManager, Mapping] = None):
        if isinstance(value, BlockManager):
            self._data = value
        elif isinstance(value, Mapping):
            self._data = BlockManager.from_dict(value)
        elif isinstance(value, Sequence):
            if not isinstance(value[0], Mapping):
                raise ValueError(
                    "Cannot set DataPanel `data` to a Sequence containing object of "
                    f" type {type(value[0])}. Must be a Sequence of Mapping."
                )
            gen = (list(x.keys()) for x in value)
            columns = lib.fast_unique_multiple_list_gen(gen)
            data = lib.dicts_to_array(value, columns=columns)
            # Assert all columns are the same length
            data = {column: list(data[:, idx]) for idx, column in enumerate(columns)}
            self._data = BlockManager.from_dict(data)
        elif value is None:
            self._data = BlockManager()
        else:
            raise ValueError(
                f"Cannot set DataPanel `data` to object of type {type(value)}."
            )

    @data.setter
    def data(self, value):
        self._set_data(value)

    @property
    def columns(self):
        """Column names in the DataPanel."""
        return list(self.data.keys())

    @property
    def nrows(self):
        """Number of rows in the DataPanel."""
        if self.ncols == 0:
            return 0
        return self.data.nrows

    @property
    def ncols(self):
        """Number of rows in the DataPanel."""
        return self.data.ncols

    @property
    def shape(self):
        """Shape of the DataPanel (num_rows, num_columns)."""
        return self.nrows, self.ncols

    def add_column(
        self, name: str, data: AbstractColumn.Columnable, overwrite=False
    ) -> None:
        """Add a column to the DataPanel."""

        assert isinstance(
            name, str
        ), f"Column name must of type `str`, not `{type(name)}`."

        assert (name not in self.columns) or overwrite, (
            f"Column with name `{name}` already exists, "
            f"set `overwrite=True` to overwrite."
        )

        if name in self.columns:
            self.remove_column(name)

        column = AbstractColumn.from_data(data)

        assert len(column) == len(self), (
            f"`add_column` failed. "
            f"Values length {len(column)} != dataset length {len(self)}."
        )

        # Add the column
        self.data[name] = column

        logger.info(f"Added column `{name}` with length `{len(column)}`.")

    def remove_column(self, column: str) -> None:
        """Remove a column from the dataset."""
        assert column in self.columns, f"Column `{column}` does not exist."

        # Remove the column
        del self.data[column]

        logger.info(f"Removed column `{column}`.")

    @capture_provenance(capture_args=["axis"])
    def append(
        self,
        dp: DataPanel,
        axis: Union[str, int] = "rows",
        suffixes: Tuple[str] = None,
        overwrite: bool = False,
    ) -> DataPanel:
        """Append a batch of data to the dataset.

        `example_or_batch` must have the same columns as the dataset
        (regardless of what columns are visible).
        """
        return meerkat.concat(
            [self, dp], axis=axis, suffixes=suffixes, overwrite=overwrite
        )

    def _add_index(self):
        """Add an index to the dataset."""
        self.add_column("index", [str(i) for i in range(len(self))])

    def head(self, n: int = 5) -> DataPanel:
        """Get the first `n` examples of the DataPanel."""
        return self.lz[:n]

    def tail(self, n: int = 5) -> DataPanel:
        """Get the last `n` examples of the DataPanel."""
        return self.lz[-n:]

    def _get(self, index, materialize: bool = False):
        if isinstance(index, str):
            # str index => column selection (AbstractColumn)
            if index in self.columns:
                return self.data[index]
            raise KeyError(f"Column `{index}` does not exist.")

        elif isinstance(index, int):
            # int index => single row (dict)
            return {
                k: self.data[k]._get(index, materialize=materialize)
                for k in self.columns
            }

        # cases where `index` returns a datapanel
        index_type = None
        if isinstance(index, slice):
            # slice index => multiple row selection (DataPanel)
            index_type = "row"

        elif (isinstance(index, tuple) or isinstance(index, list)) and len(index):
            # tuple or list index => multiple row selection (DataPanel)
            if isinstance(index[0], str):
                index_type = "column"
            else:
                index_type = "row"

        elif isinstance(index, np.ndarray):
            if len(index.shape) != 1:
                raise ValueError(
                    "Index must have 1 axis, not {}".format(len(index.shape))
                )
            # numpy array index => multiple row selection (DataPanel)
            index_type = "row"

        elif torch.is_tensor(index):
            if len(index.shape) != 1:
                raise ValueError(
                    "Index must have 1 axis, not {}".format(len(index.shape))
                )
            # torch tensor index => multiple row selection (DataPanel)
            index_type = "row"

        elif isinstance(index, pd.Series):
            index_type = "row"

        elif isinstance(index, AbstractColumn):
            # column index => multiple row selection (DataPanel)
            index_type = "row"

        else:
            raise TypeError("Invalid index type: {}".format(type(index)))

        if index_type == "column":
            if not set(index).issubset(self.columns):
                missing_cols = set(index) - set(self.columns)
                raise KeyError(f"DataPanel does not have columns {missing_cols}")

            dp = self._clone(data=self.data[index])
            return dp
        elif index_type == "row":  # pragma: no cover
            return self._clone(
                data=self.data.apply("_get", index=index, materialize=materialize)
            )

    # @capture_provenance(capture_args=[])
    def __getitem__(self, index):
        return self._get(index, materialize=True)

    def __setitem__(self, index, value):
        self.add_column(name=index, data=value, overwrite=True)

    @property
    def has_index(self) -> bool:
        """Check if the dataset has an index column."""
        if self.columns:
            return "index" in self.columns
        # Just return True if the dataset is empty
        return True

    def consolidate(self):
        self.data.consolidate()

    @classmethod
    def from_huggingface(cls, *args, **kwargs):
        """Load a Huggingface dataset as a DataPanel.

        Use this to replace `datasets.load_dataset`, so

        >>> dict_of_datasets = datasets.load_dataset('boolq')

        becomes

        >>> dict_of_datapanels = DataPanel.from_huggingface('boolq')
        """
        # Load the dataset
        dataset = datasets.load_dataset(*args, **kwargs)

        if isinstance(dataset, dict):
            return dict(
                map(
                    lambda t: (t[0], cls(t[1])),
                    dataset.items(),
                )
            )
        else:
            return cls(dataset)

    @classmethod
    @capture_provenance()
    def from_jsonl(
        cls,
        json_path: str,
    ) -> DataPanel:
        """Load a dataset from a .jsonl file on disk, where each line of the
        json file consists of a single example."""
        with open(json_path) as f:
            data = {k: [] for k in json.loads(f.readline())}
        # Load the .jsonl file
        with open(json_path) as f:
            for line in f:
                line = json.loads(line)
                for k in data:
                    data[k].append(line[k])

        return cls(
            data=data,
        )

    @classmethod
    @capture_provenance()
    def from_batch(
        cls,
        batch: Batch,
    ) -> DataPanel:
        """Convert a batch to a Dataset."""
        return cls(batch)

    @classmethod
    @capture_provenance()
    def from_batches(
        cls,
        batches: Sequence[Batch],
    ) -> DataPanel:
        """Convert a list of batches to a dataset."""

        return cls.from_batch(
            tz.merge_with(
                tz.compose(list, tz.concat),
                *batches,
            ),
        )

    @classmethod
    @capture_provenance()
    def from_dict(
        cls,
        d: Dict,
    ) -> DataPanel:
        """Convert a dictionary to a dataset.

        Alias for Dataset.from_batch(..).
        """
        return cls.from_batch(
            batch=d,
        )

    @classmethod
    @capture_provenance()
    def from_pandas(
        cls,
        df: pd.DataFrame,
    ):
        """Create a Dataset from a pandas DataFrame."""
        # column names must be str in meerkat
        df = df.rename(mapper=str, axis="columns")
        return cls.from_batch(
            df.to_dict("series"),
        )

    @classmethod
    @capture_provenance()
    def from_arrow(
        cls,
        table: pa.Table,
    ):
        """Create a Dataset from a pandas DataFrame."""
        from meerkat.block.arrow_block import ArrowBlock
        from meerkat.columns.arrow_column import ArrowArrayColumn

        block_views = ArrowBlock.from_block_data(table)
        return cls.from_batch(
            {view.block_index: ArrowArrayColumn(view) for view in block_views}
        )

    @classmethod
    @capture_provenance(capture_args=["filepath"])
    def from_csv(cls, filepath: str, *args, **kwargs):
        """Create a Dataset from a csv file.

        Args:
            filepath (str): The file path or buffer to load from.
                Same as :func:`pandas.read_csv`.
            *args: Argument list for :func:`pandas.read_csv`.
            **kwargs: Keyword arguments for :func:`pandas.read_csv`.

        Returns:
            DataPanel: The constructed datapanel.
        """
        return cls.from_pandas(pd.read_csv(filepath, *args, **kwargs))

    @classmethod
    @capture_provenance()
    def from_feather(
        cls,
        path: str,
    ):
        """Create a Dataset from a feather file."""
        return cls.from_batch(
            pd.read_feather(path).to_dict("list"),
        )

    @capture_provenance()
    def to_pandas(self) -> pd.DataFrame:
        """Convert a Dataset to a pandas DataFrame."""
        return pd.DataFrame(
            {
                name: column.to_pandas().reset_index(drop=True)
                for name, column in self.items()
            }
        )

    def to_jsonl(self, path: str) -> None:
        """Save a Dataset to a jsonl file."""
        with jsonlines.open(path, mode="w") as writer:
            for example in self:
                writer.write(example)

    def _get_collate_fns(self, columns: Iterable[str] = None):
        columns = self.data.keys() if columns is None else columns
        return {name: self.data[name].collate for name in columns}

    def _collate(self, batch: List):
        batch = tz.merge_with(list, *batch)
        column_to_collate = self._get_collate_fns(batch.keys())
        new_batch = {}
        for name, values in batch.items():
            new_batch[name] = column_to_collate[name](values)
        dp = self._clone(data=new_batch)
        return dp

    @staticmethod
    def _convert_to_batch_fn(
        function: Callable, with_indices: bool, materialize: bool = True, **kwargs
    ) -> callable:
        return convert_to_batch_fn(
            function=function,
            with_indices=with_indices,
            materialize=materialize,
            **kwargs,
        )

    def batch(
        self,
        batch_size: int = 1,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        materialize: bool = True,
        shuffle: bool = False,
        *args,
        **kwargs,
    ):
        """Batch the dataset.
        TODO:

        Args:
            batch_size: integer batch size
            drop_last_batch: drop the last batch if its smaller than batch_size

        Returns:
            batches of data
        """
        cell_columns, batch_columns = [], []
        from meerkat.columns.lambda_column import LambdaColumn

        for name, column in self.items():
            if isinstance(column, (CellColumn, LambdaColumn)) and materialize:
                cell_columns.append(name)
            else:
                batch_columns.append(name)

        indices = np.arange(len(self))

        if shuffle:
            indices = np.random.permutation(indices)

        if batch_columns:
            batch_indices = []
            for i in range(0, len(self), batch_size):
                if drop_last_batch and i + batch_size > len(self):
                    continue
                batch_indices.append(indices[i : i + batch_size])
            batch_dl = torch.utils.data.DataLoader(
                self[batch_columns] if materialize else self[batch_columns].lz,
                sampler=batch_indices,
                batch_size=None,
                batch_sampler=None,
                drop_last=drop_last_batch,
                num_workers=num_workers,
                *args,
                **kwargs,
            )

        if cell_columns:
            dp = self[cell_columns] if not shuffle else self[cell_columns].lz[indices]
            cell_dl = torch.utils.data.DataLoader(
                dp if materialize else dp.lz,
                batch_size=batch_size,
                collate_fn=self._collate,
                drop_last=drop_last_batch,
                num_workers=num_workers,
                *args,
                **kwargs,
            )

        if batch_columns and cell_columns:
            for cell_batch, batch_batch in zip(cell_dl, batch_dl):
                yield self._clone(data={**cell_batch.data, **batch_batch.data})
        elif batch_columns:
            for batch_batch in batch_dl:
                yield batch_batch
        elif cell_columns:
            for cell_batch in cell_dl:
                yield cell_batch

    @capture_provenance(capture_args=["with_indices"])
    def update(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        remove_columns: Optional[List[str]] = None,
        num_workers: int = 0,
        output_type: Union[type, Dict[str, type]] = None,
        mmap: bool = False,
        mmap_path: str = None,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> DataPanel:
        """Update the columns of the dataset."""
        # TODO(karan): make this fn go faster
        # most of the time is spent on the merge, speed it up further

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return self

        # Get some information about the function
        dp = self[input_columns] if input_columns is not None else self
        function_properties = dp._inspect_function(
            function, with_indices, is_batched_fn, materialize=materialize, **kwargs
        )
        assert (
            function_properties.dict_output
        ), f"`function` {function} must return dict."

        if not is_batched_fn:
            # Convert to a batch function
            function = convert_to_batch_fn(
                function, with_indices=with_indices, materialize=materialize, **kwargs
            )
            logger.info(f"Converting `function` {function} to batched function.")

        # Update always returns a new dataset
        logger.info("Running update, a new dataset will be returned.")

        # Copy the ._data dict with a reference to the actual columns
        new_dp = self.view()

        # Calculate the values for the new columns using a .map()
        output = new_dp.map(
            function=function,
            with_indices=with_indices,
            is_batched_fn=True,
            batch_size=batch_size,
            num_workers=num_workers,
            output_type=output_type,
            input_columns=input_columns,
            mmap=mmap,
            mmap_path=mmap_path,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )

        # Add new columns for the update
        for col, vals in output.data.items():
            if col == "index":
                continue
            new_dp.add_column(col, vals, overwrite=True)

        # Remove columns
        if remove_columns:
            for col in remove_columns:
                new_dp.remove_column(col)
            logger.info(f"Removed columns {remove_columns}.")

        return new_dp

    @capture_provenance()
    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        output_type: Union[type, Dict[str, type]] = None,
        mmap: bool = False,
        mmap_path: str = None,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> Optional[Union[Dict, List, AbstractColumn]]:
        input_columns = self.columns if input_columns is None else input_columns
        dp = self[input_columns]
        return super(DataPanel, dp).map(
            function=function,
            with_indices=with_indices,
            is_batched_fn=is_batched_fn,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            num_workers=num_workers,
            output_type=output_type,
            mmap=mmap,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )

    @capture_provenance(capture_args=["function"])
    def filter(
        self,
        function: Optional[Callable] = None,
        with_indices=False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> Optional[DataPanel]:
        """Filter operation on the DataPanel."""

        # Return if `self` has no examples
        if not len(self):
            logger.info("DataPanel empty, returning None.")
            return None

        # Get some information about the function
        dp = self[input_columns] if input_columns is not None else self
        function_properties = dp._inspect_function(
            function,
            with_indices,
            is_batched_fn=is_batched_fn,
            materialize=materialize,
            **kwargs,
        )
        assert function_properties.bool_output, "function must return boolean."

        # Map to get the boolean outputs and indices
        logger.info("Running `filter`, a new DataPanel will be returned.")
        outputs = self.map(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            is_batched_fn=is_batched_fn,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            num_workers=num_workers,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )
        indices = np.where(outputs)[0]

        # filter returns a new datapanel
        return self.lz[indices]

    def merge(
        self,
        right: meerkat.DataPanel,
        how: str = "inner",
        on: Union[str, List[str]] = None,
        left_on: Union[str, List[str]] = None,
        right_on: Union[str, List[str]] = None,
        sort: bool = False,
        suffixes: Sequence[str] = ("_x", "_y"),
        validate=None,
        keep_indexes: bool = False,
    ):
        from meerkat import merge

        return merge(
            self,
            right,
            how=how,
            on=on,
            left_on=left_on,
            right_on=right_on,
            sort=sort,
            suffixes=suffixes,
            validate=validate,
            keep_indexes=keep_indexes,
        )

    def items(self):
        for name in self.columns:
            yield name, self.data[name]

    def keys(self):
        return self.columns

    def values(self):
        for name in self.columns:
            yield self.data[name]

    @classmethod
    def read(
        cls,
        path: str,
        *args,
        **kwargs,
    ) -> DataPanel:
        """Load a DataPanel stored on disk."""

        # Load the metadata
        metadata = dict(
            yaml.load(open(os.path.join(path, "meta.yaml")), Loader=MeerkatLoader)
        )

        state = dill.load(open(os.path.join(path, "state.dill"), "rb"))
        dp = cls.__new__(cls)
        dp._set_state(state)

        # Load the the manager
        mgr_dir = os.path.join(path, "mgr")
        if os.path.exists(mgr_dir):
            data = BlockManager.read(mgr_dir, **kwargs)
        else:
            # backwards compatability to pre-manager datapanels
            data = {
                name: dtype.read(os.path.join(path, "columns", name), *args, **kwargs)
                for name, dtype in metadata["column_dtypes"].items()
            }

        dp._set_data(data)

        return dp

    def write(
        self,
        path: str,
    ) -> None:
        """Save a DataPanel to disk."""
        # Make all the directories to the path
        os.makedirs(path, exist_ok=True)

        # Get the DataPanel state
        state = self._get_state()

        # Get the metadata
        metadata = {
            "dtype": type(self),
            "column_dtypes": {name: type(col) for name, col in self.data.items()},
            "len": len(self),
        }

        # write the block manager
        mgr_dir = os.path.join(path, "mgr")
        self.data.write(mgr_dir)

        # Write the state
        state_path = os.path.join(path, "state.dill")
        dill.dump(state, open(state_path, "wb"))

        # Save the metadata as a yaml file
        metadata_path = os.path.join(path, "meta.yaml")
        yaml.dump(metadata, open(metadata_path, "w"))

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {}

    def _view_data(self) -> object:
        return self.data.view()

    def _copy_data(self) -> object:
        return self.data.copy()

    def __finalize__(self, *args, **kwargs):
        return self



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Hashable, Mapping, Sequence, Tuple, Union

import yaml

from meerkat.errors import ConsolidationError

# an index into a block that specifies where a column's data lives in the block
BlockIndex = Union[int, slice, str]


if TYPE_CHECKING:
    from meerkat.block.ref import BlockRef


@dataclass
class BlockView:
    block_index: BlockIndex
    block: AbstractBlock

    @property
    def data(self):
        return self.block._get_data(self.block_index)


class AbstractBlock:
    def __init__(self, *args, **kwargs):
        super(AbstractBlock, self).__init__(*args, **kwargs)

    def __getitem__(self, index: BlockIndex) -> BlockView:
        return BlockView(block_index=index, block=self)

    def _get_data(self, index: BlockIndex) -> object:
        """Must return view of the underlying data."""
        raise NotImplementedError()

    @property
    def signature(self) -> Hashable:
        raise NotImplementedError

    @classmethod
    def from_column_data(cls, data: object) -> Tuple[AbstractBlock, BlockView]:
        raise NotImplementedError()

    @classmethod
    def from_block_data(cls, data: object) -> Tuple[AbstractBlock, BlockView]:
        raise NotImplementedError()

    @classmethod
    def consolidate(
        cls, block_refs: Sequence[BlockRef]
    ) -> Tuple[AbstractBlock, Mapping[str, BlockIndex]]:
        if len(block_refs) == 0:
            raise ConsolidationError("Must pass at least 1 BlockRef to consolidate.")

        if len({ref.block.signature for ref in block_refs}) != 1:
            raise ConsolidationError(
                "Can only consolidate blocks with matching signatures."
            )

        return cls._consolidate(block_refs=block_refs)

    @classmethod
    def _consolidate(cls, block_refs: Sequence[BlockRef]) -> BlockRef:
        raise NotImplementedError

    def _get(self, index, block_ref: BlockRef) -> Union[BlockRef, dict]:
        raise NotImplementedError

    @property
    def is_mmap(self):
        return False

    def write(self, path: str, *args, **kwargs):
        os.makedirs(path, exist_ok=True)
        self._write_data(path, *args, **kwargs)
        metadata = {"klass": type(self)}
        metadata_path = os.path.join(path, "meta.yaml")
        yaml.dump(metadata, open(metadata_path, "w"))

    @classmethod
    def read(cls, path: str, *args, **kwargs):
        assert os.path.exists(path), f"`path` {path} does not exist."
        metadata_path = os.path.join(path, "meta.yaml")
        metadata = dict(yaml.load(open(metadata_path), Loader=yaml.FullLoader))

        block_class = metadata["klass"]
        data = block_class._read_data(path, *args, **kwargs)
        return block_class(data)



# we only include a subset of the fields in the dicom metadata


TABLES = {
    "cxr_records": {
        "table": "physionet-data.mimic_cxr.record_list",
        "fields": [
            "study_id",
            "subject_id",
            "dicom_id",
            # use tuples to specify alias
            ("path", "dicom_path"),
        ],
    },
    "cxr_studies": {
        "table": "physionet-data.mimic_cxr.study_list",
        "fields": [
            ("path", "report_path"),
        ],
    },
    "labels": {
        "table": "physionet-data.mimic_cxr.chexpert",
        "fields": [
            "Atelectasis",
            "Cardiomegaly",
            "Consolidation",
            "Edema",
            "Enlarged_Cardiomediastinum",
            "Fracture",
            "Lung_Lesion",
            "Lung_Opacity",
            "No_Finding",
            "Pleural_Effusion",
            "Pleural_Other",
            "Pneumonia",
            "Pneumothorax",
            "Support_Devices",
        ],
    },
    "dicom_meta": {
        "table": "physionet-data.mimic_cxr.dicom_metadata_string",
        "fields": [
            "dicom",
            "StudyDate",
            "ImageType",
            "TableType",
            "DistanceSourceToDetector",
            "DistanceSourceToPatient",
            "Exposure",
            "ExposureTime",
            "XRayTubeCurrent",
            "FieldOfViewRotation",
            "FieldOfViewOrigin",
            "FieldOfViewHorizontalFlip",
            "ViewPosition",
            "PatientOrientation",
            "BurnedInAnnotation",
            "RequestingService",
            "DetectorPrimaryAngle",
            "DetectorElementPhysicalSize",
        ],
    },
    "patients": {
        "table": "physionet-data.mimic_core.patients",
        "fields": ["gender", "anchor_age", "anchor_year", "anchor_year_group", "dod"],
    },
    "admit": {
        "table": "physionet-data.mimic_core.admissions",
        "fields": [
            "hadm_id",
            "admittime",
            "dischtime",
            "deathtime",
            "admission_type",
            "admission_location",
            "discharge_location",
            "insurance",
            "language",
            "marital_status",
            "ethnicity",
            "edregtime",
            "edouttime",
            "hospital_expire_flag",
        ],
    },
}



class MaterializationMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def lz(self):
        return _LazyIndexer(self)


class _LazyIndexer:
    def __init__(self, obj: object):
        self.obj = obj

    def __getitem__(self, index):
        return self.obj._get(index, materialize=False)

    def __len__(self):
        return len(self.obj)



from __future__ import annotations

import os
from typing import Sequence, Set

import pyarrow as pa
import torch
from pyarrow.compute import equal

from meerkat.block.abstract import BlockView
from meerkat.block.arrow_block import ArrowBlock
from meerkat.columns.abstract import AbstractColumn
from meerkat.errors import ImmutableError


class ArrowArrayColumn(
    AbstractColumn,
):

    block_class: type = ArrowBlock

    def __init__(
        self,
        data: Sequence,
        *args,
        **kwargs,
    ):
        if isinstance(data, BlockView):
            if not isinstance(data.block, ArrowBlock):
                raise ValueError(
                    "ArrowArrayColumn can only be initialized with ArrowBlock."
                )
        elif not isinstance(data, (pa.Array, pa.ChunkedArray)):
            data = pa.array(data)

        super(ArrowArrayColumn, self).__init__(data=data, *args, **kwargs)

    def _get(self, index, materialize: bool = True):
        index = ArrowBlock._convert_index(index)

        if isinstance(index, slice) or isinstance(index, int):
            data = self._data[index]
        elif index.dtype == bool:
            data = self._data.filter(pa.array(index))
        else:
            data = self._data.take(index)

        if self._is_batch_index(index):
            return self._clone(data=data)
        else:
            return data

    def _set(self, index, value):
        raise ImmutableError("ArrowArrayColumn is immutable.")

    def _repr_cell(self, index) -> object:
        return self.data[index]

    def is_equal(self, other: AbstractColumn) -> bool:
        if other.__class__ != self.__class__:
            return False
        return equal(self.data, other.data)

    @classmethod
    def _state_keys(cls) -> Set:
        return super()._state_keys()

    def _write_data(self, path):
        table = pa.Table.from_arrays([self.data], names=["0"])
        ArrowBlock._write_table(os.path.join(path, "data.arrow"), table)

    @staticmethod
    def _read_data(path, mmap=False):
        table = ArrowBlock._read_table(os.path.join(path, "data.arrow"), mmap=mmap)
        return table["0"]

    @classmethod
    def concat(cls, columns: Sequence[ArrowArrayColumn]):
        data = pa.concat_arrays([c.data for c in columns])
        return columns[0]._clone(data=data)

    def to_numpy(self):
        return self.data.to_numpy(zero_copy_only=False)

    def to_tensor(self):
        return torch.tensor(self.data.to_numpy())

    def to_pandas(self):
        return self.data.to_pandas()



import random

import numpy as np
import pandas as pd
from PIL import Image
from torchvision.datasets.cifar import CIFAR10

import meerkat as mk


def get_torchvision_dataset(dataset_name, download_dir, is_train):
    return NotImplemented
    # dataset = torchvision.datasets.__dict__[dataset_name.upper()](
    #     root=download_dir,
    #     train=is_train,
    #     download=True,
    # )


def get_all_cifar10_splits(
    download_dir,
    frac_val=0.0,
    transforms=None,
    seed=42,
):
    """Load a CIFAR10 split as a Meerkat DataPanel.

    Args:
        download_dir: download directory
        frac_val: fraction of training set to use for validation
        transforms: torchvision tranforms to apply to train, val and test
        seed: random seed

    Returns:
        DataPanels for train, val and test containing columns
            `raw_image`, `image` and `label`
    """
    # Validation fraction should  be between 0 and 1
    assert 0 <= frac_val <= 1, "frac_val must be between 0 and 1"

    # Load the train and test datasets
    train_dataset = CIFAR10(
        root=download_dir,
        train=True,
        download=True,
    )

    test_dataset = CIFAR10(
        root=download_dir,
        train=False,
        download=True,
    )

    # Build the train and test DataPanels
    train_dp = mk.DataPanel(
        {
            "raw_image": train_dataset.data,
            "label": mk.TensorColumn(train_dataset.targets),
        }
    )
    train_dp["split"] = np.array(["train"] * len(train_dp))

    test_dp = mk.DataPanel(
        {
            "raw_image": test_dataset.data,
            "label": mk.TensorColumn(test_dataset.targets),
        }
    )
    test_dp["split"] = np.array(["test"] * len(test_dp))

    if frac_val > 1e-4:
        # sample indices for splitting off val
        random.seed(seed)
        val_indices = set(
            random.sample(
                range(len(train_dp)),
                int(frac_val * len(train_dp)),
            )
        )

        train_dp["split"] = pd.Series(
            ["train" if i not in val_indices else "val" for i in range(len(train_dp))]
        )

    # Split up the train dataset into train and val
    val_dp = train_dp.lz[train_dp["split"] == "val"]
    train_dp = train_dp.lz[train_dp["split"] != "val"]

    def _transform(x):
        """Convert to PIL image and then apply the transforms."""
        x = Image.fromarray(x)
        if transforms:
            return transforms(x)
        return x

    train_dp["image"] = mk.LambdaColumn(train_dp["raw_image"], _transform)
    val_dp["image"] = mk.LambdaColumn(val_dp["raw_image"], _transform)
    test_dp["image"] = mk.LambdaColumn(test_dp["raw_image"], _transform)

    return {
        "train": train_dp,
        "val": val_dp,
        "test": test_dp,
    }


def get_cifar10(
    download_dir, is_train=True, frac_val=0.0, transforms=None, download: bool = True
):
    """Load CIFAR10 as a Meerkat DataPanel.

    Args:
        download_dir: download directory
        is_train: load train set
        frac_val: fraction of training set to use for validation
        transforms: torchvision tranforms to apply

    Returns:
        a DataPanel containing columns `raw_image`, `image` and `label`
    """
    dataset = CIFAR10(
        root=download_dir,
        train=is_train,
        download=download,
    )

    dp = mk.DataPanel(
        {
            "raw_image": dataset.data,
            "label": mk.TensorColumn(dataset.targets),
        }
    )

    def _transform(x):
        """Convert to PIL image and then apply the transforms."""
        x = Image.fromarray(x)
        if transforms:
            return transforms(x)
        return x

    dp["image"] = mk.LambdaColumn(dp["raw_image"], _transform)
    dp["split"] = np.array(["train"] * len(dp))

    if is_train and frac_val > 1e-4:
        # sample indices for splitting off val
        val_indices = set(
            random.sample(
                range(len(dp)),
                int(frac_val * len(dp)),
            )
        )

        dp["split"] = np.array(
            ["train" if i not in val_indices else "val" for i in range(len(dp))]
        )

    return dp



from __future__ import annotations

import logging
import os
from typing import Collection, Sequence

from meerkat.columns.abstract import AbstractColumn
from meerkat.columns.lambda_column import LambdaCell, LambdaColumn
from meerkat.columns.pandas_column import PandasSeriesColumn
from meerkat.tools.lazy_loader import LazyLoader

folder = LazyLoader("torchvision.datasets.folder")

logger = logging.getLogger(__name__)


class ImageLoaderMixin:
    def fn(self, filepath: str):
        if self.base_dir is not None:
            filepath = os.path.join(self.base_dir, filepath)
        image = self.loader(filepath)
        if self.transform is not None:
            image = self.transform(image)
        return image


class ImageCell(ImageLoaderMixin, LambdaCell):
    def __init__(
        self,
        transform: callable = None,
        loader: callable = None,
        data: str = None,
        base_dir: str = None,
    ):
        self.loader = self.default_loader if loader is None else loader
        self.transform = transform
        self._data = data
        self.base_dir = base_dir

    def __eq__(self, other):
        return (
            (other.__class__ == self.__class__)
            and (self.data == other.data)
            and (self.transform == other.transform)
            and (self.loader == other.loader)
        )

    def __repr__(self):
        transform = getattr(self.transform, "__qualname__", repr(self.transform))
        dirs = self.data.split("/")
        short_path = ("" if len(dirs) <= 2 else ".../") + "/".join(dirs[-2:])
        return f"ImageCell({short_path}, transform={transform})"


class ImageColumn(ImageLoaderMixin, LambdaColumn):
    def __init__(
        self,
        data: Sequence[str] = None,
        transform: callable = None,
        loader: callable = None,
        base_dir: str = None,
        *args,
        **kwargs,
    ):
        if not isinstance(data, PandasSeriesColumn):
            data = PandasSeriesColumn(data)
        super(ImageColumn, self).__init__(data, *args, **kwargs)
        self.loader = self.default_loader if loader is None else loader
        self.transform = transform
        self.base_dir = base_dir

    def _create_cell(self, data: object) -> ImageCell:
        return ImageCell(
            data=data,
            loader=self.loader,
            transform=self.transform,
            base_dir=self.base_dir,
        )

    @classmethod
    def from_filepaths(
        cls,
        filepaths: Sequence[str],
        loader: callable = None,
        transform: callable = None,
        base_dir: str = None,
        *args,
        **kwargs,
    ):
        return cls(
            data=filepaths,
            loader=loader,
            transform=transform,
            base_dir=base_dir,
            *args,
            **kwargs,
        )

    @classmethod
    def default_loader(cls, *args, **kwargs):
        return folder.default_loader(*args, **kwargs)

    @classmethod
    def _state_keys(cls) -> Collection:
        return (super()._state_keys() | {"transform", "loader", "base_dir"}) - {"fn"}

    def _set_state(self, state: dict):
        state["base_dir"] = state.get("base_dir", None)  # backwards compatibility
        super()._set_state(state)

    def is_equal(self, other: AbstractColumn) -> bool:
        return (
            (other.__class__ == self.__class__)
            and (self.loader == other.loader)
            and (self.transform == other.transform)
            and self.data.is_equal(other.data)
        )



from __future__ import annotations

from typing import Dict, List

import torch

from meerkat.columns.tensor_column import TensorColumn
from meerkat.datapanel import DataPanel
from meerkat.ml.instances_column import InstancesColumn
from meerkat.ml.model import Model
from meerkat.ml.segmentation_column import SegmentationOutputColumn


class TensorModel(Model):
    def __init__(
        self,
        # identifier: str,
        model: torch.nn.Module,
        is_classifier: bool = None,
        task: str = None,
        device: str = None,
    ):

        if model is None:
            raise ValueError(
                f"A PyTorch model is required with {self.__class__.__name__}."
            )
        super(TensorModel, self).__init__(
            # identifier=identifier,
            model=model,
            is_classifier=is_classifier,
            task=task,
            device=device,
        )

        # Move the model to device
        self.to(self.device)

    def forward(self, input_batch: Dict) -> Dict:
        # Run the model on the input_batch
        with torch.no_grad():
            outputs = self.model(input_batch)

        if self.is_classifier:
            # probs and preds can be handled at ClassificationOutputColumn
            # TODO(Priya): See if there is any case where these are to be returned
            output_dict = {"logits": outputs.to("cpu")}

        elif self.task == "semantic_segmentation":
            # Output is a dict with key 'out'
            output_dict = {"logits": outputs["out"].to("cpu")}

        elif self.task == "timeseries":
            output_dict = {"preds": outputs.to("cpu")}

        elif self.task == "instance_segmentation":
            output_dict = {
                "preds": [output["instances"].to("cpu") for output in outputs]
            }

        return output_dict

    def process_batch(self, batch: DataPanel, input_columns: List[str]):

        # Convert the batch to torch.Tensor and move to device
        if self.task == "instance_segmentation":
            input_batch = batch[input_columns[0]].data
        else:
            input_batch = batch[input_columns[0]].data.to(self.device)

        return input_batch

    def semantic_segmentation(
        self,
        dataset: DataPanel,
        input_columns: List[str],
        batch_size: int = 32,
        num_classes: int = None,
    ) -> DataPanel:

        # Handles outputs for semantic_segmentation tasks

        predictions = dataset.map(
            function=self._predict,
            is_batched_fn=True,
            batch_size=batch_size,
            output_type=SegmentationOutputColumn,
            input_cols=input_columns,
        )

        # TODO(Priya): How to pass other args of SegmentationOutputColumn above?
        output_col = SegmentationOutputColumn(
            logits=predictions["logits"].data, num_classes=num_classes
        )

        output_dp = DataPanel(
            {
                "logits": output_col,
                "probs": SegmentationOutputColumn(output_col.probabilities().data),
                "preds": SegmentationOutputColumn(output_col.predictions().data),
            }
        )

        dataset.add_column("logits", output_col)
        dataset.add_column("probs", output_col.probabilities())
        dataset.add_column("preds", output_col.predictions())

        return output_dp

    def timeseries(
        self, dataset: DataPanel, input_columns: List[str], batch_size: int = 32
    ) -> DataPanel:

        # Handles outputs for timeseries

        output_dp = dataset.map(
            function=self._predict,
            is_batched_fn=True,
            batch_size=batch_size,
            output_type=TensorColumn,
            input_cols=input_columns,
        )

        dataset.add_column("preds", output_dp["preds"])

        return output_dp

    def instance_segmentation(
        self, dataset: DataPanel, input_columns: List[str], batch_size: int = 32
    ) -> DataPanel:

        # Handles outputs for instance segmentation

        output_dp = dataset.map(
            function=self._predict,
            is_batched_fn=True,
            batch_size=batch_size,
            output_type=InstancesColumn,
            input_cols=input_columns,
        )

        dataset.add_column("preds", output_dp["preds"])

        return output_dp

    def output(
        self,
        dataset: DataPanel,
        input_columns: List[str],
        batch_size: int = 32,
        num_classes: int = None,
        multi_label: bool = False,
        one_hot: bool = None,
        threshold=0.5,
    ):

        if self.is_classifier:
            return self.classification(
                dataset,
                input_columns,
                batch_size,
                num_classes,
                multi_label,
                one_hot,
                threshold,
            )
        elif self.task == "semantic_segmentation":
            return self.semantic_segmentation(
                dataset, input_columns, batch_size, num_classes
            )
        elif self.task == "timeseries":
            return self.timeseries(dataset, input_columns, batch_size)
        elif self.task == "instance_segmentation":
            return self.instance_segmentation(dataset, input_columns, batch_size)
        else:
            raise NotImplementedError



from __future__ import annotations

import io
import os
from typing import Sequence

from google.cloud import storage

from meerkat import ImageColumn
from meerkat.columns.lambda_column import LambdaCell, LambdaColumn
from meerkat.columns.pandas_column import PandasSeriesColumn


class GCSImageCell(LambdaCell):
    def __init__(
        self,
        transform: callable = None,
        loader: callable = None,
        data: str = None,
    ):
        self.loader = self.default_loader if loader is None else loader
        self.transform = transform
        self._data = data

    def fn(self, filepath: str):
        image = self.loader(filepath)
        if self.transform is not None:
            image = self.transform(image)
        return image

    def __eq__(self, other):
        return (
            (other.__class__ == self.__class__)
            and (self.data == other.data)
            and (self.transform == other.transform)
            and (self.loader == other.loader)
        )

    def __repr__(self):
        transform = getattr(self.transform, "__qualname__", repr(self.transform))
        dirs = self.data.split("/")
        short_path = ("" if len(dirs) <= 2 else ".../") + "/".join(dirs[-2:])
        return f"ImageCell({short_path}, transform={transform})"


class GCSImageColumn(ImageColumn):
    def __init__(
        self,
        blob_names: PandasSeriesColumn = None,
        bucket_name: str = None,
        project: str = None,
        transform: callable = None,
        loader: callable = None,
        writer: callable = None,
        local_dir: str = None,
        _skip_cache: bool = False,
        *args,
        **kwargs,
    ):
        super(GCSImageColumn, self).__init__(
            blob_names, transform, loader, *args, **kwargs
        )
        self.project = project
        self.bucket_name = bucket_name
        self._set_state()
        storage_client = storage.Client(project=project)

        self.bucket = storage_client.bucket(bucket_name, user_project=project)
        self.loader = (lambda x: x) if loader is None else loader
        self.writer = writer
        self.local_dir = local_dir
        self._skip_cache = _skip_cache

    def _get_formatter(self) -> callable:
        # downloading the images from gcp for every visualization is probably not
        # what we want as it makes datapanel visualization very slow
        return None

    def _create_cell(self, data: object) -> LambdaCell:
        # don't want to create a lambda
        return LambdaColumn._create_cell(self, data)

    def fn(self, blob_name: str):
        if (
            self.local_dir is not None
            and os.path.exists(os.path.join(self.local_dir, str(blob_name)))
            and not self._skip_cache
        ):
            # fetch locally if it's been cached locally
            return super(GCSImageColumn, self).fn(
                os.path.join(self.local_dir, str(blob_name))
            )
        # otherwise pull from GCP
        out = self.loader(
            io.BytesIO(self.bucket.blob(str(blob_name)).download_as_bytes())
        )

        if self.writer is not None and self.local_dir is not None:
            # cache locally if writer and local dir are both provided
            path = os.path.join(self.local_dir, str(blob_name))
            os.makedirs(os.path.dirname(path), exist_ok=True)
            self.writer(path, out)
        return out

    @classmethod
    def from_blob_names(
        cls,
        blob_names: Sequence[str],
        loader: callable = None,
        transform: callable = None,
        *args,
        **kwargs,
    ):
        if not isinstance(blob_names, PandasSeriesColumn):
            blob_names = PandasSeriesColumn(blob_names)
        return cls(
            blob_names=blob_names,
            loader=loader,
            transform=transform,
            *args,
            **kwargs,
        )

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return super()._state_keys() | {
            "bucket_name",
            "project",
            "local_dir",
            "writer",
            "_skip_cache",
        }

    @classmethod
    def _clone_keys(cls) -> set:
        # need to avoid reaccessing bucket on clone, too slow
        return {"bucket"}

    def _set_state(self, state: dict = None):
        if state is not None:
            self.__dict__.update(state)
        if state is None or "bucket" not in state:
            storage_client = storage.Client(project=self.project)
            self.bucket = storage_client.bucket(
                self.bucket_name, user_project=self.project
            )



import functools
import os
from typing import Any, List, Optional, Sequence

from fvcore.common.registry import Registry as _Registry
from tabulate import tabulate

from meerkat.config import ContribOptions
from meerkat.datapanel import DataPanel


class Registry(_Registry):
    """Extension of fvcore's registry that supports aliases."""

    _ALIAS_KEYWORDS = ("_aliases", "_ALIASES")

    def __init__(self, name: str):

        super().__init__(name=name)

        self._metadata_map = {}

    def get(
        self, name: str, dataset_dir: str = None, download: bool = True, *args, **kwargs
    ) -> Any:
        ret = self._obj_map.get(name)
        if ret is None:
            raise KeyError(
                "No object named '{}' found in '{}' registry!".format(name, self._name)
            )
        if dataset_dir is None:
            dataset_dir = os.path.join(ContribOptions.download_dir, name)
        return ret(dataset_dir=dataset_dir, download=download, *args, **kwargs)

    def _get_aliases(self, obj_func_or_class):
        for kw in self._ALIAS_KEYWORDS:
            if hasattr(obj_func_or_class, kw):
                return getattr(obj_func_or_class, kw)
        return []

    def register(
        self, obj: object = None, aliases: Sequence[str] = None
    ) -> Optional[object]:
        if obj is None:
            # used as a decorator
            def deco(func_or_class: object, aliases=None) -> object:
                name = func_or_class.__name__  # pyre-ignore
                self._do_register(name, func_or_class)
                if aliases is None:
                    aliases = self._get_aliases(func_or_class)
                if not isinstance(aliases, (list, tuple, set)):
                    aliases = [aliases]
                for alias in aliases:
                    self._do_register(alias, func_or_class)
                return func_or_class

            kwargs = {"aliases": aliases}
            if any(v is not None for v in kwargs.values()):
                return functools.partial(deco, **kwargs)
            else:
                return deco

        name = obj.__name__  # pyre-ignore
        self._do_register(name, obj)
        if aliases is None:
            aliases = self._get_aliases(obj)
        for alias in aliases:
            self._do_register(alias, obj)

    def _do_register(self, name: str, obj: Any, **kwargs) -> None:
        self._metadata_map[name] = {"name": name, "description": obj.__doc__, **kwargs}
        return super()._do_register(name, obj)

    @property
    def names(self) -> List[str]:
        return list(self._obj_map.keys())

    @property
    def catalog(self) -> DataPanel:
        return DataPanel(data=list(self._metadata_map.values()))

    def __repr__(self) -> str:
        table = tabulate(self._metadata_map.values(), tablefmt="fancy_grid")
        return "Registry of {}:\n".format(self._name) + table


datasets = Registry("datasets")
datasets.__doc__ = """Registry for datasets in meerkat"""


@datasets.register()
def cifar10(dataset_dir: str = None, download: bool = True):
    """[summary]"""
    from .torchvision import get_cifar10

    return get_cifar10(download_dir=dataset_dir, download=download)


@datasets.register()
def imagenet(dataset_dir: str = None, download: bool = True):
    from .imagenet import build_imagenet_dps

    return build_imagenet_dps(dataset_dir=dataset_dir, download=download)


@datasets.register()
def imagenette(dataset_dir: str = None, download: bool = True):
    from .imagenette import build_imagenette_dp

    return build_imagenette_dp(dataset_dir=dataset_dir, download=download)


@datasets.register()
def celeba(dataset_dir: str = None, download: bool = True):
    from .celeba import get_celeba

    return get_celeba(dataset_dir=dataset_dir, download=download)



from __future__ import annotations

from meerkat.cells.abstract import AbstractCell
from meerkat.tools.lazy_loader import LazyLoader

spacy = LazyLoader("spacy")
spacy_attrs = LazyLoader("spacy.attrs")
spacy_tokens = LazyLoader("spacy.tokens")


class SpacyCell(AbstractCell):
    def __init__(
        self,
        doc: spacy_tokens.Doc,
        *args,
        **kwargs,
    ):
        super(SpacyCell, self).__init__(*args, **kwargs)

        # Put some data into the cell
        self.doc = doc

    def default_loader(self, *args, **kwargs):
        return self

    def get(self, *args, **kwargs):
        return self.doc

    def get_state(self):
        attrs = [name for name in spacy_attrs.NAMES if name != "HEAD"]
        arr = self.get().to_array(attrs)
        return {
            "arr": arr.flatten(),
            "shape": list(arr.shape),
            "words": [t.text for t in self.get()],
        }

    @classmethod
    def from_state(cls, state, nlp: spacy.language.Language):
        doc = spacy_tokens.Doc(nlp.vocab, words=state["words"])
        attrs = [name for name in spacy_attrs.NAMES if name != "HEAD"]
        return cls(doc.from_array(attrs, state["arr"].reshape(state["shape"])))

    def __getitem__(self, index):
        return self.get()[index]

    def __getattr__(self, item):
        try:
            return getattr(self.get(), item)
        except AttributeError:
            raise AttributeError(f"Attribute {item} not found.")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.get().__repr__()})"


class LazySpacyCell(AbstractCell):
    def __init__(
        self,
        text: str,
        nlp: spacy.language.Language,
        *args,
        **kwargs,
    ):
        super(LazySpacyCell, self).__init__(*args, **kwargs)

        # Put some data into the cell
        self.text = text
        self.nlp = nlp

    def default_loader(self, *args, **kwargs):
        return self

    def get(self, *args, **kwargs):
        return self.nlp(self.text)

    def get_state(self):
        return {
            "text": self.text,
        }

    @classmethod
    def from_state(cls, state, nlp: spacy.language.Language):
        return cls(text=state["text"], nlp=nlp)

    def __getitem__(self, index):
        return self.get()[index]

    def __getattr__(self, item):
        try:
            return getattr(self.get(), item)
        except AttributeError:
            raise AttributeError(f"Attribute {item} not found.")

    def __repr__(self):
        snippet = f"{self.text[:15]}..." if len(self.text) > 20 else self.text
        return f"{self.__class__.__name__}({snippet})"



import os
import subprocess
from glob import glob

import numpy as np
import pandas as pd
import torchvision.transforms as transforms
from PIL import Image

from meerkat.cells.volume import MedicalVolumeCell

GAZE_DATA_URL = "https://raw.githubusercontent.com/robustness-gym/meerkat/dev/examples/03-med_img/cxr_gaze_data.json"  # noqa: E501


def download_siim_cxr(
    dataset_dir: str,
    kaggle_username: str,
    kaggle_key: str,
    download_gaze_data: bool = True,
    include_mock_reports: bool = True,
):
    """Download the dataset from the SIIM-ACR Pneumothorax Segmentation
    challenge. https://www.kaggle.com/c/siim-acr-pneumothorax-
    segmentation/data.

    Args:
        dataset_dir (str): Path to directory where the dataset will be downloaded.
        kaggle_username (str): Your kaggle username.
        kaggle_key (str): A kaggle API key. In order to use the Kaggle’s public API, you
            must first authenticate using an API token. From the site header, click on
            your user profile picture, then on “My Account” from the dropdown menu. This
            will take you to your account settings at https://www.kaggle.com/account.
            Scroll down to the section of the page labelled API: To create a new token,
            click on the “Create New API Token” button. This will download a json file
            with a "username" and "key" field. Copy and paste the "key" field and pass
            it in as `kaggle_key`.
            Instructions copied from Kaggle API docs: https://www.kaggle.com/docs/api
        download_gaze_data (str): Download a pkl file containing eye-tracking data
            collected on a radiologist interpreting the xray.
    """
    # download and integrate gaze data
    os.environ["KAGGLE_USERNAME"] = kaggle_username
    os.environ["KAGGLE_KEY"] = kaggle_key
    subprocess.run(
        [
            "kaggle",
            "datasets",
            "download",
            "-d",
            "seesee/siim-train-test",
            "-p",
            dataset_dir,
        ]
    )
    if os.path.exists(os.path.join(dataset_dir, "siim-train-test.zip")):
        subprocess.run(
            [
                "unzip",
                "-q",
                os.path.join(dataset_dir, "siim-train-test.zip"),
                "-d",
                dataset_dir,
            ]
        )
        os.remove(os.path.join(dataset_dir, "siim-train-test.zip"))

    # get segment annotations
    segment_df = pd.read_csv(os.path.join(dataset_dir, "siim", "train-rle.csv"))
    segment_df = segment_df.rename(
        columns={"ImageId": "image_id", " EncodedPixels": "encoded_pixels"}
    )
    # there are some images that were segemented by multiple annotators, we'll just take
    # the first
    segment_df = segment_df[~segment_df.image_id.duplicated(keep="first")]

    # get binary labels for pneumothorax, any row with a "-1" for encoded pixels is
    # considered a negative
    segment_df["pmx"] = (segment_df.encoded_pixels != "-1").astype(int)

    # start building up a main dataframe with a few `merge` operations (i.e. join)
    df = segment_df

    # get filepaths for all images in the "dicom-images-train" directory
    filepaths = sorted(
        glob(os.path.join(dataset_dir, "siim", "dicom-images-train/*/*/*.dcm"))
    )
    filepath_df = pd.DataFrame(
        [
            {
                "filepath": filepath,
                "image_id": os.path.splitext(os.path.basename(filepath))[0],
            }
            for filepath in filepaths
        ]
    )

    # important to perform a left join here, because there are some images in the
    # directory without labels in `segment_df` and we only want those with labelsy
    df = df.merge(filepath_df, how="left", on="image_id")

    if download_gaze_data:
        subprocess.run(
            [
                "curl",
                GAZE_DATA_URL,
                "--output",
                os.path.join(dataset_dir, "cxr_gaze_data.json"),
            ]
        )

    if include_mock_reports:
        df["report"] = (df["pmx"] == 1).apply(_get_mock_report)

    df.to_csv(os.path.join(dataset_dir, "siim_cxr.csv"), index=False)


CXR_MEAN = 0.48865
CXR_STD = 0.24621
CXR_SIZE = 224


def cxr_transform_pil(volume: MedicalVolumeCell):
    array = volume._volume.squeeze()
    return Image.fromarray(np.uint8(array))


def cxr_transform(volume: MedicalVolumeCell):
    img = cxr_transform_pil(volume)
    img = transforms.Compose(
        [
            transforms.Resize([CXR_SIZE, CXR_SIZE]),
            transforms.ToTensor(),
            transforms.Normalize(CXR_MEAN, CXR_STD),
        ]
    )(img)
    return img.repeat([3, 1, 1])


def _get_mock_report(pmx: bool):
    state = (np.random.choice(["severe", "moderate"])) if pmx else "no"
    return np.random.choice(
        [
            (
                "Cardiac size cannot be evaluated. Large left pleural effusion is new. "
                "Small right effusion is new. The upper lungs are clear. Right lower "
                f" lobe opacities are better seen in prior CT. There is {state} "
                " pneumothorax. There are mild degenerative changes in the thoracic "
                "spine."
            ),
            (
                f"There is {state} pneumothorax. There are mild degenerative changes "
                "in the thoracic spine. The upper lungs are clear. Right lower lobe "
                "opacities are better seen in prior CT."
                "There are mild degenerative changes in the thoracic spine."
            ),
            (
                "The upper lungs are clear. Right lower lobe opacities are better "
                f"seen in prior CT. There is {state} pneumothorax. "
                "There are mild degenerative changes in the thoracic spine."
            ),
        ]
    )



import os
from pathlib import Path

from numpy.lib.format import open_memmap

from meerkat.columns.abstract import AbstractColumn
from meerkat.columns.numpy_column import NumpyArrayColumn
from meerkat.writers.abstract import AbstractWriter


class NumpyMemmapWriter(AbstractWriter):
    def __init__(
        self,
        path: str = None,
        dtype: str = "float32",
        mode: str = "r",
        shape: tuple = None,
        output_type: type = NumpyArrayColumn,
        template: AbstractColumn = None,
        *args,
        **kwargs,
    ):
        super(NumpyMemmapWriter, self).__init__(*args, **kwargs)

        # File used to store data
        self.file = None

        # Location of the pointer
        self._pointer = 0

        # If `path` is specified
        self.path = path
        self.dtype = dtype
        self.shape = shape
        if path is not None:
            self.open(path=path, dtype=dtype, mode=mode, shape=shape)

        self.output_type = output_type
        self.template = template

    def open(
        self,
        path: str,
        dtype: str = "float32",
        mode: str = "w+",
        shape: tuple = None,
    ) -> None:
        assert shape is not None, "Must specify `shape`."

        # Make all dirs to path
        os.makedirs(str(Path(path).absolute().parent), exist_ok=True)

        # Open the file as a memmap
        self.file = open_memmap(path, dtype=dtype, mode=mode, shape=shape)
        self._pointer = 0
        self.path = path
        self.shape = shape

    def write(self, arr, **kwargs) -> None:
        self.file[self._pointer : self._pointer + len(arr)] = arr
        self._pointer += len(arr)

    def flush(self):
        """Close the mmap file and reopen to release memory."""
        self.file.flush()
        self.file.base.close()
        # ‘r+’ Open existing file for reading and writing.
        self.file = open_memmap(self.file.filename, mode="r+")

    def finalize(self, *args, **kwargs) -> AbstractColumn:
        self.flush()
        data = self.file
        if self.template is not None:
            if isinstance(data, AbstractColumn):
                data = data.data
            data = self.template._clone(data=data)
        else:
            data = self.output_type(data)

        return data

    def close(self, *args, **kwargs) -> None:
        pass



from __future__ import annotations

from enum import Enum
from typing import Sequence, Union

import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from torch.distributions.categorical import Categorical

from meerkat.columns.tensor_column import TensorColumn

Columnable = Union[Sequence, np.ndarray, pd.Series, torch.Tensor]


class _ClassifierOutputType(Enum):
    LOGIT = 1, ("logit", "logits")
    PROBABILITY = 2, ("probability", "probabilities", "prob", "probs")
    PREDICTION = 3, ("prediction", "predictions", "pred", "preds")

    def __new__(cls, key_code, names):
        """
        Args:
            key_code (int): Enum value.
            extensions (tuple[str]): Extensions supported by format.
        """
        obj = object.__new__(cls)
        obj._value_ = key_code
        obj.names = names
        return obj

    @classmethod
    def get_ctype(cls, ctype):
        if isinstance(ctype, cls):
            return ctype
        if not isinstance(ctype, (int, str)):
            raise TypeError(
                f"Cannot get {cls} from ctype of type {type(ctype)}. Must be int or str"
            )

        for _ctype in cls:
            if (isinstance(ctype, int) and ctype == _ctype.value) or (
                ctype in _ctype.names
            ):
                return _ctype

        raise ValueError(f"Invalid ctype - no {cls} corresponding to `ctype={ctype}`")

    def readable_name(self):
        return self.names[0]


class ClassificationOutputColumn(TensorColumn):
    def __init__(
        self,
        logits: Columnable = None,
        probs: Columnable = None,
        preds: Columnable = None,
        num_classes: int = None,
        multi_label: bool = False,
        one_hot: bool = None,
        threshold=0.5,
        *args,
        **kwargs,
    ):
        """Classification output handler initialized by one (and only one) of
        ``logits``, ``probs``, or ``preds``.

        Args:
            logits (array-like | tensor-like): Event probabilities (unnormalized).
            probs (array-like | tensor-like): Event probabilities.
            preds (array-like | tensor-like): Predictions. Can either be one-hot
                encoded (set ``one_hot=True``) or array-like of categorical labels.
            num_classes (int): The number of categories.
            multi_label (bool): If ``True``, examples can have multiple categories.
            one_hot (bool): If ``True``, ``preds`` (if specified) is considered to
                be a one-hot encoded matrix. If ``True`` and ``multi_label=False``,
                ``preds`` will be auto-converted into categorical labels.
            threshold (float): The threshold for binarizing multi-label probabilities
                to predictions. This is not relevant if ``multi_label=False``.
        """
        valid_data = [
            (x, ctype)
            for x, ctype in [
                (logits, _ClassifierOutputType.LOGIT),
                (probs, _ClassifierOutputType.PROBABILITY),
                (preds, _ClassifierOutputType.PREDICTION),
            ]
            if x is not None
        ]

        if len(valid_data) == 0:
            raise ValueError("Must specify one of `logits`, `probs`, or `preds`")

        if len(valid_data) > 1:
            raise ValueError(
                "Only one of `logits`, `probs`, or `preds` should be specified"
            )

        data, ctype = valid_data[0]

        if isinstance(data, TensorColumn):
            data = data.data  # unwrap tensor out of TensorColumn

        if ctype == _ClassifierOutputType.PREDICTION and not multi_label and one_hot:
            # Convert one-hot single-label into categorical labels
            data = torch.argmax(data)
            one_hot = False

        self._ctype = _ClassifierOutputType.get_ctype(ctype)
        self.num_classes = num_classes
        self.multi_label = multi_label
        self.one_hot = one_hot
        self.threshold = threshold

        super().__init__(data=data, *args, **kwargs)

    def _set_data(self, data):

        if self.num_classes is None:
            if self._ctype in (
                _ClassifierOutputType.LOGIT,
                _ClassifierOutputType.PROBABILITY,
            ):
                self.num_classes = data.shape[1]
            if self._ctype == _ClassifierOutputType.PREDICTION:
                if self.multi_label:
                    if self.one_hot is False or data.ndim < 2 or not _is_binary(data):
                        raise ValueError(
                            "Multi-label predictions must be one-hot encoded "
                            "and have at least 2 dimensions - (N, C, ...)"
                        )
                    self.one_hot = True
                self.num_classes = (
                    data.shape[1] if self.multi_label else torch.max(data) + 1
                )
        super(ClassificationOutputColumn, self)._set_data(data)

    def logits(self) -> ClassificationOutputColumn:
        if self._ctype in (
            _ClassifierOutputType.PROBABILITY,
            _ClassifierOutputType.PREDICTION,
        ):
            raise ValueError(
                f"Cannot convert from {self._ctype.readable_name()} to logits"
            )

        return self.view()

    def probabilities(self) -> ClassificationOutputColumn:
        if self._ctype == _ClassifierOutputType.LOGIT:
            probs = (
                torch.sigmoid(self.data)
                if self.multi_label
                else torch.softmax(self.data, dim=1)
            )
            return ClassificationOutputColumn(
                probs=probs,
                num_classes=self.num_classes,
                multi_label=self.multi_label,
                threshold=self.threshold,
            )

        if self._ctype == _ClassifierOutputType.PROBABILITY:
            return self.view()

        raise ValueError(f"Cannot convert from {self._ctype.readable_name()} to logits")

    probs = probabilities

    def predictions(self) -> ClassificationOutputColumn:
        """Compute predictions."""
        if self._ctype == _ClassifierOutputType.PREDICTION:
            return self.view()

        probs = self.probabilities().data
        preds = (
            (probs >= self.threshold).type(torch.uint8)
            if self.multi_label
            else torch.argmax(probs, dim=1)
        )
        return ClassificationOutputColumn(
            preds=preds,
            num_classes=self.num_classes,
            multi_label=self.multi_label,
            threshold=self.threshold,
            # multi-label predictions are always one hot encoded.
            one_hot=self.multi_label,
        )

    preds = predictions

    def bincount(self) -> TensorColumn:
        """Compute the count (cardinality) for each category.

        Categories which are not available will have a count of 0.

        If ``self.multi_label=True``, the bincount will include the
        total number of times the category is seen. If an example is marked
        as 2 categories, the bincount will increase the count for both categories.
        Note, this means the sum of the number of classes can be more than
        the number of examples ``N``.

        Returns:
            torch.Tensor: A 1D tensor of length ``self.num_classes``.
        """
        preds = self.predictions().data
        if (self.multi_label and preds.ndim > 2) or (
            not self.multi_label and preds.ndim > 1
        ):
            raise ValueError(
                "mode for multi-dimensional tensors is not currently supported"
            )
        if preds.ndim == 1:
            return TensorColumn(torch.bincount(preds, minlength=self.num_classes))
        else:
            out = preds.sum(dim=0)
            if len(out) < self.num_classes:
                pad = (0,) * (2 * (out.ndim - 1)) + (0, self.num_classes - len(out))
                out = F.pad(out, pad=pad)
            return TensorColumn(out)

    def mode(self):
        count = self.bincount()
        return torch.argmax(count.data)

    def entropy(self) -> TensorColumn:
        """Compute the entropy for each example.

        If ``self.multi_label`` is True, each category is treated as a binary
        classification problem. There will be an entropy calculation for each category
        as well. For example, if the probabilities are of shape ``(N, C)``, there will
        be ``NxC`` entropy values.

        In the multi-dimensional case, this returns the entropy for each element.
        For example, if the probabilities are of shape ``(N, C, A, B)``, there will
        be ``NxAxB`` entropy values.

        Returns:
            TensorColumn: Tensor of entropies
        """
        probs = self.probabilities().data
        if self.multi_label:
            probs = torch.stack([probs, 1 - probs], dim=-1)
        elif probs.ndim > 2:
            # make channels last
            probs = probs.transpose((0,) + tuple(range(2, probs.ndim)) + (1,))
        return TensorColumn(Categorical(probs=probs).entropy())

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return super()._state_keys() | {
            "_ctype",
            "num_classes",
            "multi_label",
            "one_hot",
            "threshold",
        }


def _is_binary(tensor: torch.Tensor):
    return torch.all(
        (tensor == tensor.type(torch.uint8)) & (tensor >= 0) & (tensor <= 0)
    )



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Hashable, Sequence, Tuple, Union

import numpy as np
import torch

from meerkat.block.ref import BlockRef
from meerkat.columns.numpy_column import NumpyArrayColumn
from meerkat.errors import ConsolidationError

from .abstract import AbstractBlock, BlockIndex, BlockView


class TensorBlock(AbstractBlock):
    @dataclass(eq=True, frozen=True)
    class Signature:
        device: torch.device
        dtype: torch.dtype
        nrows: int
        shape: Tuple[int]
        klass: type

    def __init__(self, data, *args, **kwargs):
        super(TensorBlock, self).__init__(*args, **kwargs)
        if len(data.shape) <= 1:
            raise ValueError(
                "Cannot create a `TensorBlock` from data with less than 2 axes."
            )
        self.data = data

    @property
    def signature(self) -> Hashable:
        return self.Signature(
            klass=TensorBlock,
            device=self.data.device,
            nrows=self.data.shape[0],
            shape=self.data.shape[2:],
            dtype=self.data.dtype,
        )

    def _get_data(self, index: BlockIndex) -> torch.Tensor:
        return self.data[:, index]

    @classmethod
    def from_column_data(cls, data: torch.Tensor) -> Tuple[TensorBlock, BlockView]:
        """[summary]

        Args:
            data (np.ndarray): [description]
            names (Sequence[str]): [description]

        Raises:
            ValueError: [description]

        Returns:
            Tuple[NumpyBlock, Mapping[str, BlockIndex]]: [description]
        """
        if len(data.shape) == 1:
            data = torch.unsqueeze(data, dim=1)
            block_index = 0
        elif data.shape[1] == 1:
            block_index = slice(0, 1)
        else:
            block_index = slice(0, data.shape[1])

        block = cls(data)
        return BlockView(block_index=block_index, block=block)

    @classmethod
    def _consolidate(
        cls,
        block_refs: Sequence[BlockRef],
    ) -> BlockRef:
        offset = 0
        new_indices = {}
        columns = {}
        to_concat = []
        for block_ref in block_refs:
            for name, col in block_ref.items():
                # keep track of all the columns in the block_refs
                if name in columns:
                    raise ConsolidationError(
                        "Cannot consolidate two block refs containing the same column."
                    )
                columns[name] = col

                # add block and compute new indices
                block_index = col._block_index
                if isinstance(block_index, slice):
                    block_view = col._block.data[:, block_index]
                    new_indices[name] = slice(
                        # need to update slice offset and remove step
                        offset,
                        block_view.shape[1] + offset,
                        1,
                    )
                elif isinstance(block_index, int):
                    # keep block axis
                    block_view = col._block.data[:, block_index : block_index + 1]
                    new_indices[name] = offset
                to_concat.append(block_view)
                offset += block_view.shape[1]

        block = cls(torch.cat(to_concat, dim=1))

        # create columns
        new_columns = {
            name: columns[name]._clone(data=block[block_index])
            for name, block_index in new_indices.items()
        }
        return BlockRef(block=block, columns=new_columns)

    @staticmethod
    def _convert_index(index):
        if isinstance(index, NumpyArrayColumn) and index.data.dtype == np.bool_:
            # needed to silence torch deprecation warning
            # DeprecationWarning: In future, it will be an error for 'np.bool_' scalars
            # to be interpreted as an index
            return torch.as_tensor(index.data)

        from meerkat.columns.pandas_column import PandasSeriesColumn

        if (
            isinstance(index, PandasSeriesColumn)
            and index.data.values.dtype == np.bool_
        ):
            # needed to silence torch deprecation warning
            # DeprecationWarning: In future, it will be an error for 'np.bool_' scalars
            # to be interpreted as an index
            return torch.as_tensor(index.data.values)

        from meerkat.columns.tensor_column import TensorColumn

        if isinstance(index, TensorColumn):
            # need to convert to numpy for boolean indexing
            return index.data
        return index

    def _get(
        self, index, block_ref: BlockRef, materialize: bool = True
    ) -> Union[BlockRef, dict]:

        index = self._convert_index(index)
        # TODO: check if they're trying to index more than just the row dimension
        data = self.data[index]
        if isinstance(index, int):
            # if indexing a single row, we do not return a block manager, just a dict
            return {
                name: data[col._block_index] for name, col in block_ref.columns.items()
            }
        block = self.__class__(data)
        columns = {
            name: col._clone(data=block[col._block_index])
            for name, col in block_ref.columns.items()
        }
        # note that the new block may share memory with the old block
        return BlockRef(block=block, columns=columns)

    def _write_data(self, path: str):
        torch.save(self.data, os.path.join(path, "data.pt"))

    @staticmethod
    def _read_data(path: str, mmap: bool = False):
        return torch.load(os.path.join(path, "data.pt"))



from __future__ import annotations

from typing import Dict, List, Optional

import cytoolz as tz
import torch

from meerkat.columns.list_column import ListColumn
from meerkat.datapanel import DataPanel
from meerkat.ml.model import Model
from meerkat.tools.lazy_loader import LazyLoader

AutoTokenizer = LazyLoader("transformers.AutoTokenizer")


class HuggingfaceModel(Model):
    def __init__(
        self,
        identifier: str,
        model,
        tokenizer: Optional[AutoTokenizer] = None,
        is_classifier: bool = None,
        task: str = None,
        device: str = None,
    ):

        if model is None:
            raise ValueError(
                f"A HuggingFace model is required with {self.__class__.__name__}."
            )

        super(HuggingfaceModel, self).__init__(
            # identifier=identifier,
            model=model,
            is_classifier=is_classifier,
            task=task,
            device=device,
        )

        self.tokenizer = tokenizer
        if tokenizer is None:
            # Load the tokenizer
            self.tokenizer = AutoTokenizer.from_pretrained(self.identifier)

        # Move the model to device
        self.to(self.device)

    def forward(self, input_batch: Dict) -> Dict:

        if self.is_classifier:
            # Run the model on the input_batch
            with torch.no_grad():
                outputs = self.model(**input_batch)

            # probs and preds can be handled at ClassificationOutputColumn
            # TODO(Priya): See if there is any case where these are to be returned
            # Logits are present at the 0th index
            output_dict = {"logits": outputs[0].to("cpu")}

        else:
            # TODO (Priya): Support for only summarization right now.
            with torch.no_grad():
                summary_token_ids = self.model.generate(**input_batch)
                summaries = [
                    self.tokenizer.decode(
                        token_id_list,
                        skip_special_tokens=True,
                        clean_up_tokenization_spaces=False,
                    )
                    for token_id_list in summary_token_ids
                ]
                output_dict = {"preds": summaries}

        return output_dict

    def encode_batch(self, batch: DataPanel, columns: List[str], **kwargs):
        # TODO(karan): Automatically writing this encoder for a variety of tasks
        return self.tokenizer(
            *[list(batch[key]) for key in columns],
            truncation=True,
            padding=True,
            **kwargs,
        )

    def process_batch(self, batch: DataPanel, input_columns: List[str]):

        # Tokenize the batch
        input_batch = self.encode_batch(batch=batch, columns=input_columns)

        # Convert the batch to torch.Tensor
        input_batch = tz.valmap(
            lambda v: torch.tensor(v).to(device=self.device), input_batch
        )

        # Return the converted batch
        return input_batch

    def summarization(
        self, dataset: DataPanel, input_columns: List[str], batch_size: int = 32
    ) -> DataPanel:

        output_dp = dataset.map(
            function=self._predict,
            is_batched_fn=True,
            batch_size=batch_size,
            output_type=ListColumn,
            input_cols=input_columns,
        )

        dataset.add_column("preds", output_dp["preds"])

        return output_dp

    def output(
        self,
        dataset: DataPanel,
        input_columns: List[str],
        batch_size: int = 32,
        num_classes: int = None,
        multi_label: bool = False,
        one_hot: bool = None,
        threshold=0.5,
    ):
        # TODO(Priya): The separate functions can be merged later
        if self.is_classifier:
            return self.classification(
                dataset,
                input_columns,
                batch_size,
                num_classes,
                multi_label,
                one_hot,
                threshold,
            )
        elif self.task == "summarization":
            return self.summarization(dataset, input_columns, batch_size)
        else:
            raise NotImplementedError

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return super()._state_keys() | {
            "identifier",
            "model",
            "tokenizer",
            "is_classifier",
            "task",
            "device",
        }



from meerkat.block.abstract import BlockView


class BlockableMixin:
    def __init__(self, *args, **kwargs):
        super(BlockableMixin, self).__init__(*args, **kwargs)

    block_class: type = None

    @classmethod
    def is_blockable(cls):
        return cls.block_class is not None

    def _unpack_block_view(self, data):
        if isinstance(data, BlockView):
            self._block = data.block
            self._block_index = data.block_index
            data = data.data
        else:
            block_view: BlockView = self.block_class.from_column_data(data)
            self._block, self._block_index = block_view.block, block_view.block_index
        return data

    def _pack_block_view(self):
        return BlockView(block_index=self._block_index, block=self._block)



import abc


class AbstractWriter(abc.ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def __str__(self):
        return f"{self.__class__.__name__}"

    @abc.abstractmethod
    def open(self, *args, **kwargs) -> None:
        return NotImplemented

    @abc.abstractmethod
    def write(self, data, *args, **kwargs) -> None:
        return NotImplemented

    @abc.abstractmethod
    def flush(self, *args, **kwargs) -> None:
        return NotImplemented

    @abc.abstractmethod
    def close(self, *args, **kwargs) -> None:
        return NotImplemented

    @abc.abstractmethod
    def finalize(self, *args, **kwargs) -> None:
        return NotImplemented



from itertools import combinations
from typing import Sequence, Tuple, Union

import cytoolz as tz

from meerkat import DataPanel
from meerkat.columns.abstract import AbstractColumn
from meerkat.errors import ConcatError
from meerkat.provenance import capture_provenance


@capture_provenance(capture_args=["axis"])
def concat(
    objs: Union[Sequence[DataPanel], Sequence[AbstractColumn]],
    axis: Union[str, int] = "rows",
    suffixes: Tuple[str] = None,
    overwrite: bool = False,
) -> Union[DataPanel, AbstractColumn]:
    """Concatenate a sequence of columns or a sequence of `DataPanel`s. If
    sequence is empty, returns an empty `DataPanel`.

    - If concatenating columns, all columns must be of the same type (e.g. all
    `ListColumn`).
    - If concatenating `DataPanel`s along axis 0 (rows), all `DataPanel`s must have the
    same set of columns.
    - If concatenating `DataPanel`s along axis 1 (columns), all `DataPanel`s must have
    the same length and cannot have any of the same column names.

    Args:
        objs (Union[Sequence[DataPanel], Sequence[AbstractColumn]]): sequence of columns
            or DataPanels.
        axis (Union[str, int]): The axis along which to concatenate. Ignored if
            concatenating columns.

    Returns:
        Union[DataPanel, AbstractColumn]: concatenated DataPanel or column
    """
    if len(objs) == 0:
        return DataPanel()

    if not all([type(objs[0]) == type(obj) for obj in objs[1:]]):
        raise ConcatError("All objects passed to concat must be of same type.")

    if isinstance(objs[0], DataPanel):
        if axis == 0 or axis == "rows":
            # append new rows
            columns = set(objs[0].columns)
            if not all([set(dp.columns) == columns for dp in objs]):
                raise ConcatError(
                    "Can only concatenate DataPanels along axis 0 (rows) if they have "
                    " the same set of columns names."
                )
            return objs[0].from_batch(
                {column: concat([dp[column] for dp in objs]) for column in columns}
            )
        elif axis == 1 or axis == "columns":
            # append new columns
            length = len(objs[0])
            if not all([len(dp) == length for dp in objs]):
                raise ConcatError(
                    "Can only concatenate DataPanels along axis 1 (columns) if they "
                    "have the same length."
                )

            # get all column names that appear in more than one DataPanel
            shared = set()
            for dp1, dp2 in combinations(objs, 2):
                shared |= set(dp1.columns) & set(dp2.columns)

            # TODO (sabri): I removed the index column for now to address
            # https://github.com/robustness-gym/meerkat/issues/65, but when we refactor
            # index with https://github.com/robustness-gym/meerkat/issues/117 we should
            # take this out
            shared -= {"index"}
            if shared and not overwrite:
                if suffixes is None:
                    raise ConcatError("Must ")
                data = tz.merge(
                    {k + suffixes[idx] if k in shared else k: v for k, v in dp.items()}
                    for idx, dp in enumerate(objs)
                )
            else:
                data = tz.merge(dict(dp.items()) for dp in objs)

            return objs[0]._clone(data=data)
        else:
            raise ConcatError(f"Invalid axis `{axis}` passed to concat.")
    elif isinstance(objs[0], AbstractColumn):
        # use the concat method of the column
        return objs[0].concat(objs)
    else:
        raise ConcatError(
            "Must pass a sequence of datapanels or a sequence of columns to concat."
        )



from __future__ import annotations

from pathlib import Path
from typing import Callable, Dict, Sequence, Union

from meerkat.cells.abstract import AbstractCell
from meerkat.mixins.cloneable import StateClass
from meerkat.mixins.file import PathLikeType, PathsMixin
from meerkat.tools.lazy_loader import LazyLoader

pydicom = LazyLoader("pydicom")
dosma = LazyLoader("dosma")
dosma_core_io_format_io = LazyLoader("dosma.core.io.format_io")

try:
    # Mapping from pydicom types to python types
    _PYDICOM_TO_PYTHON = {
        pydicom.valuerep.DSdecimal: float,
        pydicom.valuerep.DSfloat: float,
        pydicom.valuerep.IS: int,
        pydicom.valuerep.PersonName: str,
        pydicom.multival.MultiValue: list,
        pydicom.sequence.Sequence: list,
    }
except ImportError:
    _PYDICOM_TO_PYTHON = {}


class MedicalVolumeCell(PathsMixin, AbstractCell):
    """Interface for loading medical volume data.

    Examples:

        # Specify xray dicoms with default orientation ``("SI", "AP")``:
        >>> cell = MedicalVolumeCell("/path/to/xray.dcm",
        loader=DicomReader(group_by=None, default_ornt=("SI", "AP"))

        # Load multi-echo MRI volumes
        >>> cell = MedicalVolumeCell("/path/to/mri/scan/dir",
        loader=DicomReader(group_by="EchoNumbers"))
    """

    def __init__(
        self,
        paths: Union[PathLikeType, Sequence[PathLikeType]],
        loader: Callable = None,
        transform: Callable = None,
        cache_metadata: bool = False,
        *args,
        **kwargs,
    ):
        super(MedicalVolumeCell, self).__init__(paths=paths, *args, **kwargs)
        self._metadata = None
        self.transform: Callable = transform
        self.cache_metadata = cache_metadata
        self.loader = self.default_loader(self.paths) if loader is None else loader

    def __getitem__(self, index):
        image = self.get()
        return image[index]

    def __str__(self):
        return f"{self.__class__.__name__}({self.paths})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.paths})"

    @classmethod
    def _unroll_path(cls, paths: Sequence[Path]):
        if len(paths) == 1:
            return paths[0]
        return paths

    @classmethod
    def default_loader(cls, paths: Sequence[Path], *args, **kwargs):
        paths = cls._unroll_path(paths)
        return dosma.get_reader(dosma.ImageDataFormat.get_image_data_format(paths))

    def get(self, *args, cache_metadata: bool = None, **kwargs):
        image = self.loader(self._unroll_path(self.paths))
        if cache_metadata is None:
            cache_metadata = self.cache_metadata
        # DOSMA returns a list of MedicalVolumes by default.
        # RG overrides this functinality  - if only one MedicalVolume
        # is returned, unpack that volume from the list.
        if isinstance(image, (list, tuple)) and len(image) == 1:
            image = image[0]
        if self._metadata is None and cache_metadata:
            _img = image[0] if isinstance(image, (list, tuple)) else image
            headers = _img.headers(flatten=True)
            self._metadata = self._prune_metadata(headers[0]) if headers else None
        if self.transform is not None:
            image = self.transform(image)
        return image

    def _prune_metadata(self, metadata):
        """Prunes the metadata to avoid keeping large fields.

        We delete the PixelData field, because that data is a duplicate
        of the image.
        """
        del metadata["PixelData"]
        return metadata

    def get_metadata(
        self,
        ignore_bytes: bool = False,
        readable: bool = False,
        as_raw_type: bool = False,
        force_load: bool = False,
    ) -> Dict:
        if self._metadata is None:
            if force_load:
                _ = self.get(cache_metadata=True)
                metadata = self._metadata
                if not self.cache_metadata:
                    self.clear_metadata()
            else:
                return None
        else:
            metadata = self._metadata

        # Raw data elements need to be decoded.
        metadata = {
            k: metadata[k] if hasattr(v, "is_raw") and v.is_raw else v
            for k, v in metadata.items()
        }

        if ignore_bytes:
            metadata = {
                k: v for k, v in metadata.items() if not isinstance(v.value, bytes)
            }
        if readable:
            metadata = {v.name: v for v in metadata.values()}
        if as_raw_type:
            metadata = {
                k: (
                    _PYDICOM_TO_PYTHON[type(v.value)](v.value)
                    if type(v.value) in _PYDICOM_TO_PYTHON
                    else v.value
                )
                for k, v in metadata.items()
            }

        return metadata

    def clear_metadata(self):
        self._metadata = None

    def get_state(self):
        # Check if the loader is a `DataReader` from `dosma`
        is_dosma_reader = isinstance(self.loader, dosma_core_io_format_io.DataReader)

        loader = StateClass(
            klass=type(self.loader) if is_dosma_reader else None,
            state=self.loader.state_dict() if is_dosma_reader else self.loader,
        )

        return {
            "paths": self.paths,
            "loader": loader,
            "transform": self.transform,
        }

    @classmethod
    def from_state(cls, state, *args, **kwargs):
        # Unpack the state
        loader_state = state["loader"]

        if loader_state.klass is None:
            loader = loader_state.state
        else:
            loader = loader_state.klass()
            loader.load_state_dict(loader_state.state)

        return cls(
            state["paths"],
            loader=loader,
            transform=state["transform"],
        )



from dataclasses import dataclass

from meerkat.mixins.blockable import BlockableMixin
from meerkat.provenance import ProvenanceMixin


@dataclass
class StateClass:
    """An internal class to store the state of an object alongside its
    associated class."""

    klass: type
    state: object


class CloneableMixin:
    def __init__(self, *args, **kwargs):
        super(CloneableMixin, self).__init__(*args, **kwargs)

    @classmethod
    def _state_keys(cls) -> set:
        """"""
        raise NotImplementedError()

    @classmethod
    def _clone_keys(cls) -> set:
        return set()

    def copy(self, **kwargs) -> object:
        new_data = self._copy_data()
        return self._clone(data=new_data)

    def view(self) -> object:
        return self._clone()

    def _clone(self, data: object = None):
        if data is None:
            if isinstance(self, BlockableMixin) and self.is_blockable():
                data = self._pack_block_view()
            else:
                data = self._view_data()

        state = self._get_state(clone=True)

        obj = self.__class__.__new__(self.__class__)
        obj._set_state(state)
        obj._set_data(data)

        if isinstance(self, ProvenanceMixin):
            # need to create a node for the object
            obj._init_node()

        return obj

    def _copy_data(self) -> object:
        raise NotImplementedError

    def _view_data(self) -> object:
        return self.data

    def _get_state(self, clone: bool = False) -> dict:
        state = {key: getattr(self, key) for key in self._state_keys()}
        if clone:
            state.update({key: getattr(self, key) for key in self._clone_keys()})
        return state

    def _set_state(self, state: dict):
        self.__dict__.update(state)



import logging
from typing import Callable, Dict, Mapping, Optional, Union

import numpy as np
from tqdm.auto import tqdm

from meerkat.provenance import capture_provenance

logger = logging.getLogger(__name__)


class MappableMixin:
    def __init__(self, *args, **kwargs):
        super(MappableMixin, self).__init__(*args, **kwargs)

    @capture_provenance()
    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        drop_last_batch: bool = False,
        num_workers: Optional[int] = 0,
        output_type: Union[type, Dict[str, type]] = None,
        materialize: bool = True,
        pbar: bool = False,
        mmap: bool = False,
        mmap_path: str = None,
        flush_size: int = None,
        **kwargs,
    ):
        # TODO (sabri): add materialize?
        from meerkat.columns.abstract import AbstractColumn
        from meerkat.datapanel import DataPanel

        """Map a function over the elements of the column."""

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return None

        if not is_batched_fn:
            # Convert to a batch function
            function = self._convert_to_batch_fn(
                function, with_indices=with_indices, materialize=materialize, **kwargs
            )
            is_batched_fn = True
            logger.info(f"Converting `function` {function} to a batched function.")

        # Run the map
        logger.info("Running `map`, the dataset will be left unchanged.")
        for i, batch in tqdm(
            enumerate(
                self.batch(
                    batch_size=batch_size,
                    drop_last_batch=drop_last_batch,
                    num_workers=num_workers,
                    materialize=materialize
                    # TODO: collate=batched was commented out in list_column
                )
            ),
            total=(len(self) // batch_size)
            + int(not drop_last_batch and len(self) % batch_size != 0),
            disable=not pbar,
        ):
            # Calculate the start and end indexes for the batch
            start_index = i * batch_size
            end_index = min(len(self), (i + 1) * batch_size)

            # Use the first batch for setup
            if i == 0:
                # Get some information about the function
                function_properties = self._inspect_function(
                    function,
                    with_indices,
                    is_batched_fn,
                    batch,
                    range(start_index, end_index),
                    materialize=materialize,
                    **kwargs,
                )

                # Pull out information
                output = function_properties.output
                dtype = function_properties.output_dtype
                is_mapping = isinstance(output, Mapping)
                is_type_mapping = isinstance(output_type, Mapping)

                if not is_mapping and is_type_mapping:
                    raise ValueError(
                        "output_type is a mapping but function output is not a mapping"
                    )

                writers = {}
                for key, curr_output in (
                    output.items() if is_mapping else [("0", output)]
                ):
                    curr_output_type = (
                        type(AbstractColumn.from_data(curr_output))
                        if output_type is None
                        or (is_type_mapping and key not in output_type.keys())
                        else output_type[key]
                        if is_type_mapping
                        else output_type
                    )

                    writer = curr_output_type.get_writer(
                        mmap=mmap,
                        template=(
                            curr_output.copy()
                            if isinstance(curr_output, AbstractColumn)
                            else None
                        ),
                    )

                    # Setup for writing to a certain output column
                    # TODO: support optionally memmapping only some columns
                    if mmap:
                        if not hasattr(curr_output, "shape"):
                            curr_output = np.array(curr_output)

                        # Assumes first dimension of output is the batch dimension.
                        shape = (len(self), *curr_output.shape[1:])

                        # Construct the mmap file path
                        if mmap_path is None:
                            mmap_path = self.logdir / f"{hash(function)}" / key
                        # Open the output writer
                        writer.open(str(mmap_path), dtype, shape=shape)
                    else:
                        # Create an empty dict or list for the outputs
                        writer.open()
                    writers[key] = writer

            else:
                # Run `function` on the batch
                output = (
                    function(
                        batch,
                        range(i * batch_size, min(len(self), (i + 1) * batch_size)),
                        **kwargs,
                    )
                    if with_indices
                    else function(batch, **kwargs)
                )

            # Append the output
            if output is not None:
                if isinstance(output, Mapping):
                    if set(output.keys()) != set(writers.keys()):
                        raise ValueError(
                            "Map function must return same keys for each batch."
                        )
                    for k, writer in writers.items():
                        writer.write(output[k])
                else:
                    writers["0"].write(output)

            # intermittently flush
            if flush_size is not None and ((i + 1) % flush_size == 0):
                for writer in writers.values():
                    writer.flush()

        # Check if we are returning a special output type
        outputs = {key: writer.finalize() for key, writer in writers.items()}

        if not is_mapping:
            outputs = outputs["0"]
        else:
            # TODO (arjundd): This is duck type. We should probably make this
            # class signature explicit.
            outputs = (
                self._clone(data=outputs)
                if isinstance(self, DataPanel)
                else DataPanel.from_batch(outputs)
            )
            outputs._visible_columns = None
        return outputs



from __future__ import annotations

import abc
import logging
import pathlib
import reprlib
from copy import copy
from typing import Any, Callable, List, Optional, Sequence, Union

import numpy as np
import pandas as pd
import torch

import meerkat.config
from meerkat.mixins.blockable import BlockableMixin
from meerkat.mixins.cloneable import CloneableMixin
from meerkat.mixins.collate import CollateMixin
from meerkat.mixins.inspect_fn import FunctionInspectorMixin
from meerkat.mixins.io import ColumnIOMixin
from meerkat.mixins.lambdable import LambdaMixin
from meerkat.mixins.mapping import MappableMixin
from meerkat.mixins.materialize import MaterializationMixin
from meerkat.provenance import ProvenanceMixin, capture_provenance
from meerkat.tools.utils import convert_to_batch_column_fn

logger = logging.getLogger(__name__)


class AbstractColumn(
    BlockableMixin,
    CloneableMixin,
    CollateMixin,
    ColumnIOMixin,
    FunctionInspectorMixin,
    LambdaMixin,
    MappableMixin,
    MaterializationMixin,
    ProvenanceMixin,
    abc.ABC,
):
    """An abstract class for Meerkat columns."""

    _data: Sequence = None

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "meerkat/"

    # Create a directory
    logdir.mkdir(parents=True, exist_ok=True)

    def __init__(
        self,
        data: Sequence = None,
        collate_fn: Callable = None,
        *args,
        **kwargs,
    ):
        # Assign to data
        self._set_data(data)

        super(AbstractColumn, self).__init__(
            collate_fn=collate_fn,
            *args,
            **kwargs,
        )

        # Log creation
        logger.info(f"Created `{self.__class__.__name__}` with {len(self)} rows.")

    def __repr__(self):
        return f"{self.__class__.__name__}({reprlib.repr(self.data)})"

    def __str__(self):
        return f"{self.__class__.__name__}({reprlib.repr(self.data)})"

    def streamlit(self):
        return self._repr_pandas_()

    def _set_data(self, data):
        if self.is_blockable():
            data = self._unpack_block_view(data)
        self._data = data

    @property
    def data(self):
        """Get the underlying data."""
        return self._data

    @data.setter
    def data(self, value):
        self._set_data(value)

    @property
    def metadata(self):
        return {}

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {"_collate_fn"}

    def _get_cell(self, index: int, materialize: bool = True) -> Any:
        """Get a single cell from the column.

        Args:
            index (int): This is an index into the ALL rows, not just visible rows. In
                other words, we assume that the index passed in has already been
                remapped via `_remap_index`, if `self.visible_rows` is not `None`.
            materialize (bool, optional): Materialize and return the object. This
                argument is used by subclasses of `AbstractColumn` that hold data in an
                unmaterialized format. Defaults to False.
        """
        return self._data[index]

    def _get_batch(
        self, indices: np.ndarray, materialize: bool = True
    ) -> AbstractColumn:
        """Get a batch of cells from the column.

        Args:
            index (int): This is an index into the ALL rows, not just visible rows. In
                other words, we assume that the index passed in has already been
                remapped via `_remap_index`, if `self.visible_rows` is not `None`.
            materialize (bool, optional): Materialize and return the object. This
                argument is used by subclasses of `AbstractColumn` that hold data in an
                unmaterialized format. Defaults to False.
        """
        if materialize:
            return self.collate(
                [self._get_cell(int(i), materialize=materialize) for i in indices]
            )

        else:
            return self.collate(
                [self._get_cell(int(i), materialize=materialize) for i in indices]
            )

    def _get(self, index, materialize: bool = True, _data: np.ndarray = None):
        index = self._translate_index(index)
        if isinstance(index, int):
            if _data is None:
                _data = self._get_cell(index, materialize=materialize)
            return _data

        elif isinstance(index, np.ndarray):
            # support for blocks
            if _data is None:
                _data = self._get_batch(index, materialize=materialize)
            return self._clone(data=_data)

    # @capture_provenance()
    def __getitem__(self, index):
        return self._get(index, materialize=True)

    def _set_cell(self, index, value):
        self._data[index] = value

    def _set_batch(self, indices: np.ndarray, values):
        for index, value in zip(indices, values):
            self._set_cell(int(index), value)

    def _set(self, index, value):
        index = self._translate_index(index)
        if isinstance(index, int):
            self._set_cell(index, value)
        elif isinstance(index, Sequence) or isinstance(index, np.ndarray):
            self._set_batch(index, value)
        else:
            raise ValueError

    def __setitem__(self, index, value):
        self._set(index, value)

    def _is_batch_index(self, index):
        # np.ndarray indexed with a tuple of length 1 does not return an np.ndarray
        # so we match this behavior
        return not (
            isinstance(index, int) or (isinstance(index, tuple) and len(index) == 1)
        )

    def _translate_index(self, index):
        # `index` should return a single element
        if not self._is_batch_index(index):
            return index

        if isinstance(index, pd.Series):
            index = index.values

        if torch.is_tensor(index):
            index = index.numpy()

        # `index` should return a batch
        if isinstance(index, slice):
            # int or slice index => standard list slicing
            # TODO (sabri): get rid of the np.arange here, very slow for large columns
            indices = np.arange(self.full_length())[index]
        elif isinstance(index, tuple) or isinstance(index, list):
            indices = np.array(index)
        elif isinstance(index, np.ndarray):
            if len(index.shape) != 1:
                raise TypeError(
                    "`np.ndarray` index must have 1 axis, not {}".format(
                        len(index.shape)
                    )
                )
            if index.dtype == bool:
                indices = np.where(index)[0]
            else:
                return index
        elif isinstance(index, AbstractColumn):
            # TODO (sabri): get rid of the np.arange here, very slow for large columns
            indices = np.arange(self.full_length())[index]
        else:
            raise TypeError(
                "Object of type {} is not a valid index".format(type(index))
            )
        return indices

    @staticmethod
    def _convert_to_batch_fn(
        function: Callable, with_indices: bool, materialize: bool = True, **kwargs
    ) -> callable:
        return convert_to_batch_column_fn(
            function=function,
            with_indices=with_indices,
            materialize=materialize,
            **kwargs,
        )

    def __len__(self):
        return self.full_length()

    def full_length(self):
        if self._data is None:
            return 0
        return len(self._data)

    def _repr_cell_(self, index) -> object:
        raise NotImplementedError

    def _get_formatter(self) -> callable:
        return None

    def _repr_pandas_(self, max_rows: int = None) -> pd.Series:
        if max_rows is None:
            max_rows = meerkat.config.DisplayOptions.max_rows

        if len(self) > max_rows:
            col = pd.Series(
                [self._repr_cell(idx) for idx in range(max_rows // 2)]
                + [self._repr_cell(0)]
                + [
                    self._repr_cell(idx)
                    for idx in range(len(self) - max_rows // 2, len(self))
                ]
            )
        else:
            col = pd.Series([self._repr_cell(idx) for idx in range(len(self))])

        formatter = self._get_formatter()
        return col, formatter

    def _repr_html_(self, max_rows: int = None):
        # pd.Series objects do not implement _repr_html_
        if max_rows is None:
            max_rows = meerkat.config.DisplayOptions.max_rows

        if len(self) > max_rows:
            pd_index = np.concatenate(
                (
                    np.arange(max_rows // 2),
                    np.zeros(1),
                    np.arange(len(self) - max_rows // 2, len(self)),
                ),
            )
        else:
            pd_index = np.arange(len(self))

        col_name = f"({self.__class__.__name__})"
        col, formatter = self._repr_pandas_(max_rows=max_rows)
        df = col.to_frame(name=col_name)
        df = df.set_index(pd_index.astype(int))
        return df.to_html(
            max_rows=max_rows,
            formatters={col_name: formatter},
            escape=False,
        )

    @capture_provenance()
    def filter(
        self,
        function: Callable,
        with_indices=False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        drop_last_batch: bool = False,
        num_workers: Optional[int] = 0,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> Optional[AbstractColumn]:
        """Filter the elements of the column using a function."""

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning it .")
            return self

        # Get some information about the function
        function_properties = self._inspect_function(
            function,
            with_indices,
            is_batched_fn=is_batched_fn,
            materialize=materialize,
            **kwargs,
        )
        assert function_properties.bool_output, "function must return boolean."

        # Map to get the boolean outputs and indices
        logger.info("Running `filter`, a new dataset will be returned.")
        outputs = self.map(
            function=function,
            with_indices=with_indices,
            # input_columns=input_columns,
            is_batched_fn=is_batched_fn,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            num_workers=num_workers,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )
        indices = np.where(outputs)[0]
        return self.lz[indices]

    def append(self, column: AbstractColumn) -> None:
        # TODO(Sabri): implement a naive `ComposedColumn` for generic append and
        # implement specific ones for ListColumn, NumpyColumn etc.
        raise NotImplementedError

    @staticmethod
    def concat(columns: Sequence[AbstractColumn]) -> None:
        # TODO(Sabri): implement a naive `ComposedColumn` for generic append and
        # implement specific ones for ListColumn, NumpyColumn etc.
        raise NotImplementedError

    def is_equal(self, other: AbstractColumn) -> bool:
        """Tests whether two columns.

        Args:
            other (AbstractColumn): [description]
        """
        raise NotImplementedError()

    def batch(
        self,
        batch_size: int = 1,
        drop_last_batch: bool = False,
        collate: bool = True,
        num_workers: int = 0,
        materialize: bool = True,
        *args,
        **kwargs,
    ):
        """Batch the column.

        Args:
            batch_size: integer batch size
            drop_last_batch: drop the last batch if its smaller than batch_size
            collate: whether to collate the returned batches

        Returns:
            batches of data
        """
        if (
            self._get_batch.__func__ == AbstractColumn._get_batch
            and self._get.__func__ == AbstractColumn._get
        ):
            return torch.utils.data.DataLoader(
                self if materialize else self.lz,
                batch_size=batch_size,
                collate_fn=self.collate if collate else lambda x: x,
                drop_last=drop_last_batch,
                num_workers=num_workers,
                *args,
                **kwargs,
            )
        else:
            batch_indices = []
            indices = np.arange(len(self))
            for i in range(0, len(self), batch_size):
                if drop_last_batch and i + batch_size > len(self):
                    continue
                batch_indices.append(indices[i : i + batch_size])
            return torch.utils.data.DataLoader(
                self if materialize else self.lz,
                sampler=batch_indices,
                batch_size=None,
                batch_sampler=None,
                drop_last=drop_last_batch,
                num_workers=num_workers,
                *args,
                **kwargs,
            )

    @classmethod
    def get_writer(cls, mmap: bool = False, template: AbstractColumn = None):
        from meerkat.writers.concat_writer import ConcatWriter

        if mmap:
            raise ValueError("Memmapping not supported with this column type.")
        else:
            return ConcatWriter(output_type=cls, template=template)

    Columnable = Union[Sequence, np.ndarray, pd.Series, torch.Tensor]

    @classmethod
    # @capture_provenance()
    def from_data(cls, data: Union[Columnable, AbstractColumn]):
        """Convert data to a meerkat column using the appropriate Column
        type."""
        if isinstance(data, AbstractColumn):
            # TODO: Need ton make this view but should decide where to do it exactly
            return data  # .view()

        if isinstance(data, pd.Series):
            from .pandas_column import PandasSeriesColumn

            return PandasSeriesColumn(data)

        if torch.is_tensor(data):
            from .tensor_column import TensorColumn

            return TensorColumn(data)

        if isinstance(data, np.ndarray):
            from .numpy_column import NumpyArrayColumn

            return NumpyArrayColumn(data)

        if isinstance(data, Sequence):
            from ..cells.abstract import AbstractCell

            if len(data) != 0 and isinstance(data[0], AbstractCell):
                from .cell_column import CellColumn

                return CellColumn(data)

            if len(data) != 0 and isinstance(
                data[0], (int, float, bool, np.ndarray, np.generic)
            ):
                from .numpy_column import NumpyArrayColumn

                return NumpyArrayColumn(data)

            if len(data) != 0 and isinstance(data[0], str):
                from .pandas_column import PandasSeriesColumn

                return PandasSeriesColumn(data)

            if len(data) != 0 and torch.is_tensor(data[0]):
                from .tensor_column import TensorColumn

                return TensorColumn(data)

            from .list_column import ListColumn

            return ListColumn(data)
        else:
            raise ValueError(f"Cannot create column out of data of type {type(data)}")

    def head(self, n: int = 5) -> AbstractColumn:
        """Get the first `n` examples of the column."""
        return self.lz[:n]

    def tail(self, n: int = 5) -> AbstractColumn:
        """Get the last `n` examples of the column."""
        return self.lz[-n:]

    def to_pandas(self) -> pd.Series:
        return pd.Series([self.lz[int(idx)] for idx in range(len(self))])

    def _copy_data(self) -> object:
        return copy(self._data)

    def _view_data(self) -> object:
        return self._data

    @property
    def is_mmap(self):
        return False



from __future__ import annotations

import logging
from typing import Callable, Optional, Sequence

from meerkat.cells.video import VideoCell
from meerkat.columns.cell_column import CellColumn

logger = logging.getLogger(__name__)


class VideoColumn(CellColumn):
    """Interface for creating a CellColumn from VideoCell objects."""

    def __init__(self, *args, **kwargs):
        super(VideoColumn, self).__init__(*args, **kwargs)

    @classmethod
    def from_filepaths(
        cls,
        filepaths: Optional[Sequence[str]] = None,
        time_dim: Optional[int] = 1,
        # TODO: add different loaders to VideoCell
        transform: Optional[Callable] = None,
        *args,
        **kwargs,
    ):
        cells = [
            VideoCell(fp, time_dim=time_dim, transform=transform) for fp in filepaths
        ]
        return cls(
            cells=cells,
            *args,
            **kwargs,
        )



import logging
from typing import Callable

logger = logging.getLogger(__name__)


class LambdaMixin:
    def __init__(self, *args, **kwargs):
        super(LambdaMixin, self).__init__(*args, **kwargs)

    def to_lambda(self, fn: Callable = None):
        from meerkat import LambdaColumn

        return LambdaColumn(data=self, fn=fn)



from typing import Dict, List

import torch

from meerkat.datapanel import DataPanel
from meerkat.ml.embedding_column import EmbeddingColumn
from meerkat.ml.metrics import compute_metric
from meerkat.ml.prediction_column import ClassificationOutputColumn


# TODO(Priya): Move some general functions here
class Model(torch.nn.Module):
    def __init__(
        self,
        # identifier: str,
        model: torch.nn.Module,
        # evaluation_fn=None,
        is_classifier: bool = None,
        task: str = None,
        device: str = None,
    ):

        super(Model, self).__init__()

        self.model = model

        # if evaluation_fn is not None:
        #    self.evaluate = evaluation_fn

        if task is None:
            if is_classifier is None or not is_classifier:
                raise ValueError("Task is required for non-classification models.")
        else:
            is_classifier = False

        self.is_classifier = is_classifier
        self.task = task

        if not device:
            self.device = "cpu"
            if torch.cuda.is_available():
                self.device = "cuda:0"

    def forward(self, input_batch: Dict) -> Dict:
        raise NotImplementedError

    def _predict(self, batch, input_cols):
        # Use input_cols instead of input_columns to avoid naming conflict with map

        # Process the batch to prepare input
        input_batch = self.process_batch(batch, input_cols)
        # Run forward pass
        prediction_dict = self.forward(input_batch)
        return prediction_dict

    def _activation(self, batch, input_cols, activation_op):
        # Use input_cols instead of input_columns to avoid naming conflict with map

        # Process the batch
        input_batch = self.process_batch(batch, input_cols)
        # Run forward pass
        _ = self.forward(input_batch)

        activation_dict = {
            f"activation_{activation_op.target_module}": EmbeddingColumn(
                activation_op.extractor.activation.cpu().detach()
            )
        }

        return activation_dict

    def activation(
        self,
        dataset: DataPanel,
        target_module: str,  # TODO(Priya): Support multiple activation layers
        input_columns: List[str],
        batch_size=32,
    ) -> EmbeddingColumn:

        """An Operation that stores model activations in a new Embedding
        column.

        Args:
            dataset (DataPanel): the meerkat DataPanel containing the model inputs.
            target_module (str): the name of the submodule of `model` (i.e. an
                intermediate layer) that outputs the activations we'd like to extract.
                For nested submodules, specify a path separated by "." (e.g.
                `ActivationCachedOp(model, "block4.conv")`).
            input_columns (str): Column containing model inputs
        """

        # To avoid circular dependency for now.
        from meerkat.ml.activation import ActivationOp

        # Get an activation operator
        activation_op = ActivationOp(self.model, target_module, self.device)

        activations = dataset.map(
            function=self._activation,
            is_batched_fn=True,
            batch_size=batch_size,
            output_type=EmbeddingColumn,
            input_cols=input_columns,
            activation_op=activation_op,
        )

        activation_col = activations[f"activation_{activation_op.target_module}"]

        # dataset.add_column(f"activation ({target_module})", activation_col)
        return activation_col

    def classification(
        self,
        dataset: DataPanel,
        input_columns: List[str],
        batch_size: int = 32,
        num_classes: int = None,
        multi_label: bool = False,
        one_hot: bool = None,
        threshold=0.5,
    ) -> DataPanel:

        # Handles outputs for classification tasks

        predictions = dataset.map(
            function=self._predict,
            is_batched_fn=True,
            batch_size=batch_size,
            output_type=ClassificationOutputColumn,
            input_cols=input_columns,
        )

        # TODO(Priya): How to pass other args of ClassificationOutputColumn above?
        output_col = ClassificationOutputColumn(
            logits=predictions["logits"].data,
            num_classes=num_classes,
            multi_label=multi_label,
            one_hot=one_hot,
            threshold=threshold,
        )

        output_dp = DataPanel(
            {
                "logits": output_col,
                "probs": output_col.probabilities(),
                "preds": output_col.predictions(),
            }
        )

        dataset.add_column("logits", output_col)
        dataset.add_column("probs", output_col.probabilities())
        dataset.add_column("preds", output_col.predictions())

        return output_dp

    def evaluate(
        self,
        dataset: DataPanel,
        target_column: List[str],  # str?
        pred_column: List[str],  # str?
        metrics: List[str],
        num_classes: int = None,
    ):
        preds = dataset[pred_column[0]]
        labels = dataset[target_column[0]]

        if num_classes is None:
            if isinstance(preds, ClassificationOutputColumn):
                num_classes = preds.num_classes
            elif self.is_classifier:
                raise ValueError(
                    "Must specify num_classes if column type \
                        is not ClassificationOutputColumn"
                )

        evaluation_dict = {
            metric: compute_metric(metric, preds.data, labels.data, num_classes)
            for metric in metrics
        }

        return evaluation_dict

    @staticmethod
    def remap_labels(output_dict: Dict, label_map: List[int]) -> Dict:
        """Map the output labels of the model.

        Example: 3-way classificaiton, with label_map = [1, 2, 0]
        => (model label 0 -> dataset label 1, model label 1 -> dataset label 2, ...).
        """

        # Check the number of classes
        num_classes = len(label_map)

        # Remap the columns of all outputs that have # columns = num_classes
        for key in output_dict:
            if output_dict[key].shape[-1] == num_classes:
                output_dict[key] = output_dict[key][..., label_map]

        # Remap the pred key
        inverse_label_map = [
            t[1] for t in sorted([(label, i) for i, label in enumerate(label_map)])
        ]
        output_dict["pred"] = torch.tensor(inverse_label_map)[output_dict["pred"]]

        return output_dict

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {"model", "is_classifier", "task", "device"}



from __future__ import annotations

import abc
import functools
import logging
import numbers
import os
from typing import Callable, Sequence

import numpy as np
import pandas as pd
import torch
from pandas.core.accessor import CachedAccessor
from pandas.core.arrays.categorical import CategoricalAccessor
from pandas.core.dtypes.common import (
    is_categorical_dtype,
    is_datetime64_dtype,
    is_datetime64tz_dtype,
    is_period_dtype,
    is_timedelta64_dtype,
)
from pandas.core.dtypes.generic import ABCSeries
from pandas.core.indexes.accessors import (
    CombinedDatetimelikeProperties,
    DatetimeProperties,
    PeriodProperties,
    TimedeltaProperties,
)
from pandas.core.strings import StringMethods
from yaml.representer import Representer

from meerkat.block.abstract import BlockView
from meerkat.block.pandas_block import PandasBlock
from meerkat.columns.abstract import AbstractColumn

Representer.add_representer(abc.ABCMeta, Representer.represent_name)

logger = logging.getLogger(__name__)


def getattr_decorator(fn: Callable):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        out = fn(*args, **kwargs)
        if isinstance(out, pd.Series):
            return PandasSeriesColumn(out)
        elif isinstance(out, pd.DataFrame):
            from meerkat import DataPanel

            # column names must be str in meerkat
            out = out.rename(mapper=str, axis="columns")
            return DataPanel.from_pandas(out)
        else:
            return out

    return wrapper


class _ReturnColumnMixin:
    def __getattribute__(self, name):
        try:
            attr = super().__getattribute__(name)
            if isinstance(attr, Callable):
                return getattr_decorator(attr)
            elif isinstance(attr, pd.Series):
                return PandasSeriesColumn(attr)
            elif isinstance(attr, pd.DataFrame):
                from meerkat import DataPanel

                return DataPanel.from_pandas(attr)
            else:
                return attr
        except AttributeError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )


class _MeerkatStringMethods(_ReturnColumnMixin, StringMethods):
    pass


class _MeerkatDatetimeProperties(_ReturnColumnMixin, DatetimeProperties):
    pass


class _MeerkatTimedeltaProperties(_ReturnColumnMixin, TimedeltaProperties):
    pass


class _MeerkatPeriodProperties(_ReturnColumnMixin, PeriodProperties):
    pass


class _MeerkatCategoricalAccessor(_ReturnColumnMixin, CategoricalAccessor):
    pass


class _MeerkatCombinedDatetimelikeProperties(CombinedDatetimelikeProperties):
    def __new__(cls, data: pd.Series):
        # CombinedDatetimelikeProperties isn't really instantiated. Instead
        # we need to choose which parent (datetime or timedelta) is
        # appropriate. Since we're checking the dtypes anyway, we'll just
        # do all the validation here.

        if not isinstance(data, ABCSeries):
            raise TypeError(
                f"cannot convert an object of type {type(data)} to a datetimelike index"
            )

        orig = data if is_categorical_dtype(data.dtype) else None
        if orig is not None:
            data = data._constructor(
                orig.array,
                name=orig.name,
                copy=False,
                dtype=orig._values.categories.dtype,
            )

        if is_datetime64_dtype(data.dtype):
            obj = _MeerkatDatetimeProperties(data, orig)
        elif is_datetime64tz_dtype(data.dtype):
            obj = _MeerkatDatetimeProperties(data, orig)
        elif is_timedelta64_dtype(data.dtype):
            obj = _MeerkatTimedeltaProperties(data, orig)
        elif is_period_dtype(data.dtype):
            obj = _MeerkatPeriodProperties(data, orig)
        else:
            raise AttributeError("Can only use .dt accessor with datetimelike values")

        return obj


class PandasSeriesColumn(
    AbstractColumn,
    np.lib.mixins.NDArrayOperatorsMixin,
):
    block_class: type = PandasBlock

    _HANDLED_TYPES = (np.ndarray, numbers.Number, str)

    str = CachedAccessor("str", _MeerkatStringMethods)
    dt = CachedAccessor("dt", _MeerkatCombinedDatetimelikeProperties)
    cat = CachedAccessor("cat", _MeerkatCategoricalAccessor)
    # plot = CachedAccessor("plot", pandas.plotting.PlotAccessor)
    # sparse = CachedAccessor("sparse", SparseAccessor)

    def _set_data(self, data: object):
        if isinstance(data, BlockView):
            if not isinstance(data.block, PandasBlock):
                raise ValueError(
                    "Cannot create `PandasSeriesColumn` from a `BlockView` not "
                    "referencing a `PandasBlock`."
                )
        elif not isinstance(data, pd.Series):
            data = pd.Series(data)

        super(PandasSeriesColumn, self)._set_data(data)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Only support operations with instances of _HANDLED_TYPES.
            # Use ArrayLike instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle ArrayLike objects.
            if not isinstance(x, self._HANDLED_TYPES + (PandasSeriesColumn,)):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(
            x.data if isinstance(x, PandasSeriesColumn) else x for x in inputs
        )
        if out:
            kwargs["out"] = tuple(
                x.data if isinstance(x, PandasSeriesColumn) else x for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(type(self)(x) for x in result)  # pragma: no cover
        elif method == "at":
            # no return value
            return None  # pragma: no cover
        else:
            # one return value
            return type(self)(result)

    def __getattr__(self, name):
        if name == "__getstate__" or name == "__setstate__":
            # for pickle, it's important to raise an attribute error if __getstate__
            # or __setstate__ is called. Without this, pickle will use the __setstate__
            # and __getstate__ of the underlying pandas Series
            raise AttributeError()
        try:
            out = getattr(object.__getattribute__(self, "data"), name)
            if isinstance(out, Callable):
                return getattr_decorator(out)
            else:
                return out
        except AttributeError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )

    @classmethod
    def from_array(cls, data: np.ndarray, *args, **kwargs):
        return cls(data=data, *args, **kwargs)

    def _get(self, index, materialize: bool = True):
        index = self.block_class._convert_index(index)

        data = self._data.iloc[index]
        if self._is_batch_index(index):
            # only create a numpy array column
            return self._clone(data=data)
        else:
            return data

    def _set_cell(self, index, value):
        self._data.iloc[index] = value

    def _set_batch(self, indices, values):
        self._data.iloc[indices] = values

    @classmethod
    def concat(cls, columns: Sequence[PandasSeriesColumn]):
        data = pd.concat([c.data for c in columns])
        return columns[0]._clone(data=data)

    def _write_data(self, path: str) -> None:
        data_path = os.path.join(path, "data.pd")
        self.data.to_pickle(data_path)

    @staticmethod
    def _read_data(
        path: str,
    ):
        data_path = os.path.join(path, "data.pd")

        # Load in the data
        return pd.read_pickle(data_path)

    def _repr_cell(self, index) -> object:
        return self[index]

    def to_tensor(self) -> torch.Tensor:
        """Use `column.to_tensor()` instead of `torch.tensor(column)`, which is
        very slow."""
        dtype = self.data.values.dtype
        if not np.issubdtype(dtype, np.number):
            raise ValueError(
                f"Cannot convert `PandasSeriesColumn` with dtype={dtype} to tensor."
            )

        # TODO (Sabri): understand why `torch.tensor(column)` is so slow
        return torch.tensor(self.data.values)

    def is_equal(self, other: AbstractColumn) -> bool:
        if other.__class__ != self.__class__:
            return False
        return (self.data.values == other.data.values).all()

    def to_pandas(self) -> pd.Series:
        return self.data
